#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains an object that represents a Telegram ShippingOption."""

from collections.abc import Sequence
from typing import TYPE_CHECKING

from telegram._telegramobject import TelegramObject
from telegram._utils.argumentparsing import parse_sequence_arg
from telegram._utils.types import JSONDict

if TYPE_CHECKING:
    from telegram import LabeledPrice


class ShippingOption(TelegramObject):
    """This object represents one shipping option.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`id` is equal.

    Examples:
        :any:`Payment Bot <examples.paymentbot>`

    Args:
        id (:obj:`str`): Shipping option identifier.
        title (:obj:`str`): Option title.
        prices (Sequence[:class:`telegram.LabeledPrice`]): List of price portions.

            .. versionchanged:: 20.0
                |sequenceclassargs|

    Attributes:
        id (:obj:`str`): Shipping option identifier.
        title (:obj:`str`): Option title.
        prices (tuple[:class:`telegram.LabeledPrice`]): List of price portions.

            .. versionchanged:: 20.0
                |tupleclassattrs|

    """

    __slots__ = ("id", "prices", "title")

    def __init__(
        self,
        id: str,  # pylint: disable=redefined-builtin
        title: str,
        prices: Sequence["LabeledPrice"],
        *,
        api_kwargs: JSONDict | None = None,
    ):
        super().__init__(api_kwargs=api_kwargs)

        self.id: str = id
        self.title: str = title
        self.prices: tuple[LabeledPrice, ...] = parse_sequence_arg(prices)

        self._id_attrs = (self.id,)

        self._freeze()



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains a mutable mapping that keeps track of the keys that where accessed.

.. versionadded:: 20.0

Warning:
    Contents of this module are intended to be used internally by the library and *not* by the
    user. Changes to this module are not considered breaking changes and may not be documented in
    the changelog.
"""

from collections import UserDict
from collections.abc import Mapping
from typing import Final, Generic, TypeVar

from telegram._utils.defaultvalue import DEFAULT_NONE, DefaultValue

_VT = TypeVar("_VT")
_KT = TypeVar("_KT")
_T = TypeVar("_T")


class TrackingDict(UserDict, Generic[_KT, _VT]):
    """Mutable mapping that keeps track of which keys where accessed with write access.
    Read-access is not tracked.

    Note:
        * ``setdefault()`` and ``pop`` are considered writing only depending on whether the
            key is present
        * deleting values is considered writing
    """

    DELETED: Final = object()
    """Special marker indicating that an entry was deleted."""

    __slots__ = ("_write_access_keys",)

    def __init__(self) -> None:
        super().__init__()
        self._write_access_keys: set[_KT] = set()

    def __setitem__(self, key: _KT, value: _VT) -> None:
        self.__track_write(key)
        super().__setitem__(key, value)

    def __delitem__(self, key: _KT) -> None:
        self.__track_write(key)
        super().__delitem__(key)

    def __track_write(self, key: _KT | set[_KT]) -> None:
        if isinstance(key, set):
            self._write_access_keys |= key
        else:
            self._write_access_keys.add(key)

    def pop_accessed_keys(self) -> set[_KT]:
        """Returns all keys that were write-accessed since the last time this method was called."""
        out = self._write_access_keys
        self._write_access_keys = set()
        return out

    def pop_accessed_write_items(self) -> list[tuple[_KT, _VT]]:
        """
        Returns all keys & corresponding values as set of tuples that were write-accessed since
        the last time this method was called. If a key was deleted, the value will be
        :attr:`DELETED`.
        """
        keys = self.pop_accessed_keys()
        return [(key, self.get(key, self.DELETED)) for key in keys]

    def mark_as_accessed(self, key: _KT) -> None:
        """Use this method have the key returned again in the next call to
        :meth:`pop_accessed_write_items` or :meth:`pop_accessed_keys`
        """
        self._write_access_keys.add(key)

    # Override methods to track access

    def update_no_track(self, mapping: Mapping[_KT, _VT]) -> None:
        """Like ``update``, but doesn't count towards write access."""
        for key, value in mapping.items():
            self.data[key] = value

    # Mypy seems a bit inconsistent about what it wants as types for `default` and return value
    # so we just ignore a bit
    def pop(  # type: ignore[override]
        self,
        key: _KT,
        default: _VT = DEFAULT_NONE,  # type: ignore[assignment]
    ) -> _VT:
        if key in self:
            self.__track_write(key)
        if isinstance(default, DefaultValue):
            return super().pop(key)
        return super().pop(key, default=default)

    def clear(self) -> None:
        self.__track_write(set(super().keys()))
        super().clear()

    # Mypy seems a bit inconsistent about what it wants as types for `default` and return value
    # so we just ignore a bit
    def setdefault(self: "TrackingDict[_KT, _T]", key: _KT, default: _T | None = None) -> _T:
        if key in self:
            return self[key]

        self.__track_write(key)
        self[key] = default  # type: ignore[assignment]
        return default  # type: ignore[return-value]



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains an object that represents a Telegram MessageEntity."""

import copy
import datetime as dtm
import itertools
from collections.abc import Sequence
from typing import TYPE_CHECKING, Final

from telegram import constants
from telegram._telegramobject import TelegramObject
from telegram._user import User
from telegram._utils import enum
from telegram._utils.argumentparsing import de_json_optional
from telegram._utils.datetime import extract_tzinfo_from_defaults, from_timestamp
from telegram._utils.strings import TextEncoding
from telegram._utils.types import JSONDict

if TYPE_CHECKING:
    from telegram import Bot

_SEM = Sequence["MessageEntity"]


class MessageEntity(TelegramObject):
    """
    This object represents one special entity in a text message. For example, hashtags,
    usernames, URLs, etc.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`type`, :attr:`offset` and :attr:`length` are equal.

    Args:
        type (:obj:`str`): Type of the entity. Can be :attr:`MENTION` (``@username``),
            :attr:`HASHTAG` (``#hashtag`` or ``#hashtag@chatusername``), :attr:`CASHTAG` (``$USD``
            or ``USD@chatusername``), :attr:`BOT_COMMAND` (``/start@jobs_bot``), :attr:`URL`
            (``https://telegram.org``), :attr:`EMAIL` (``do-not-reply@telegram.org``),
            :attr:`PHONE_NUMBER` (``+1-212-555-0123``),
            :attr:`BOLD` (**bold text**), :attr:`ITALIC` (*italic text*), :attr:`UNDERLINE`
            (underlined text), :attr:`STRIKETHROUGH`, :attr:`SPOILER` (spoiler message),
            :attr:`BLOCKQUOTE` (block quotation), :attr:`CODE` (monowidth string), :attr:`PRE`
            (monowidth block), :attr:`TEXT_LINK` (for clickable text URLs), :attr:`TEXT_MENTION`
            (for users without usernames), :attr:`CUSTOM_EMOJI` (for inline custom emoji stickers)
            or :attr:`DATE_TIME` (for formatted date and time).

            .. versionadded:: 20.0
                Added inline custom emoji

            .. versionadded:: 20.8
                Added block quotation

            .. versionadded:: 22.7
                Added date_time
        offset (:obj:`int`): Offset in UTF-16 code units to the start of the entity.
        length (:obj:`int`): Length of the entity in UTF-16 code units.
        url (:obj:`str`, optional): For :attr:`TEXT_LINK` only, url that will be opened after
            user taps on the text.
        user (:class:`telegram.User`, optional): For :attr:`TEXT_MENTION` only, the mentioned
             user.
        language (:obj:`str`, optional): For :attr:`PRE` only, the programming language of
            the entity text.
        custom_emoji_id (:obj:`str`, optional): For :attr:`CUSTOM_EMOJI` only, unique identifier
            of the custom emoji. Use :meth:`telegram.Bot.get_custom_emoji_stickers` to get full
            information about the sticker.

            .. versionadded:: 20.0
        date_time_format (:obj:`str`, optional): For :attr:`DATE_TIME` only, the string that
            defines the formatting of the date and time. See `date-time entity formatting
            <https://core.telegram.org/bots/api#date-time-entity-formatting>`_ for more details and
            :class:`telegram.constants.MessageEntityDateTimeFormats` for all possible formats.

            .. versionadded:: 22.7
        unix_time (:class:`datetime.datetime`, optional): For :attr:`DATE_TIME` only, the time
            associated with the entity.
            |datetime_localization|

            .. versionadded:: 22.7
    Attributes:
        type (:obj:`str`): Type of the entity.  Can be :attr:`MENTION` (``@username``),
            :attr:`HASHTAG` (``#hashtag`` or ``#hashtag@chatusername``), :attr:`CASHTAG` (``$USD``
            or ``USD@chatusername``), :attr:`BOT_COMMAND` (``/start@jobs_bot``), :attr:`URL`
            (``https://telegram.org``), :attr:`EMAIL` (``do-not-reply@telegram.org``),
            :attr:`PHONE_NUMBER` (``+1-212-555-0123``),
            :attr:`BOLD` (**bold text**), :attr:`ITALIC` (*italic text*), :attr:`UNDERLINE`
            (underlined text), :attr:`STRIKETHROUGH`, :attr:`SPOILER` (spoiler message),
            :attr:`BLOCKQUOTE` (block quotation), :attr:`CODE` (monowidth string), :attr:`PRE`
            (monowidth block), :attr:`TEXT_LINK` (for clickable text URLs), :attr:`TEXT_MENTION`
            (for users without usernames), :attr:`CUSTOM_EMOJI` (for inline custom emoji stickers)
            or :attr:`DATE_TIME` (for formatted date and time).

            .. versionadded:: 20.0
                Added inline custom emoji

            .. versionadded:: 20.8
                Added block quotation

            .. versionadded:: 22.7
                Added date_time
        offset (:obj:`int`): Offset in UTF-16 code units to the start of the entity.
        length (:obj:`int`): Length of the entity in UTF-16 code units.
        url (:obj:`str`): Optional. For :attr:`TEXT_LINK` only, url that will be opened after
            user taps on the text.
        user (:class:`telegram.User`): Optional. For :attr:`TEXT_MENTION` only, the mentioned
             user.
        language (:obj:`str`): Optional. For :attr:`PRE` only, the programming language of
            the entity text.
        custom_emoji_id (:obj:`str`): Optional. For :attr:`CUSTOM_EMOJI` only, unique identifier
            of the custom emoji. Use :meth:`telegram.Bot.get_custom_emoji_stickers` to get full
            information about the sticker.

            .. versionadded:: 20.0
        date_time_format (:obj:`str`): Optional. For :attr:`DATE_TIME` only, the string that
            defines the formatting of the date and time. See `date-time entity formatting
            <https://core.telegram.org/bots/api#date-time-entity-formatting>`_ for more details and
            :class:`telegram.constants.MessageEntityDateTimeFormats` for all possible formats.

            .. versionadded:: 22.7
        unix_time (:class:`datetime.datetime`): Optional. For :attr:`DATE_TIME` only, the time
            associated with the entity.
            |datetime_localization|

            .. versionadded:: 22.7

    """

    __slots__ = (
        "custom_emoji_id",
        "date_time_format",
        "language",
        "length",
        "offset",
        "type",
        "unix_time",
        "url",
        "user",
    )

    def __init__(
        self,
        type: str,  # pylint: disable=redefined-builtin
        offset: int,
        length: int,
        url: str | None = None,
        user: User | None = None,
        language: str | None = None,
        custom_emoji_id: str | None = None,
        date_time_format: str | None = None,
        unix_time: dtm.datetime | None = None,
        *,
        api_kwargs: JSONDict | None = None,
    ):
        super().__init__(api_kwargs=api_kwargs)
        # Required
        self.type: str = enum.get_member(constants.MessageEntityType, type, type)
        self.offset: int = offset
        self.length: int = length
        # Optionals
        self.url: str | None = url
        self.user: User | None = user
        self.language: str | None = language
        self.custom_emoji_id: str | None = custom_emoji_id
        self.date_time_format: str | None = date_time_format
        self.unix_time: dtm.datetime | None = unix_time

        self._id_attrs = (self.type, self.offset, self.length)

        self._freeze()

    @classmethod
    def de_json(cls, data: JSONDict, bot: "Bot | None" = None) -> "MessageEntity":
        """See :meth:`telegram.TelegramObject.de_json`."""
        data = cls._parse_data(data)

        data["user"] = de_json_optional(data.get("user"), User, bot)

        # Get the local timezone from the bot if it has defaults
        loc_tzinfo = extract_tzinfo_from_defaults(bot)
        data["unix_time"] = from_timestamp(data.get("unix_time"), tzinfo=loc_tzinfo)

        return super().de_json(data=data, bot=bot)

    @staticmethod
    def adjust_message_entities_to_utf_16(text: str, entities: _SEM) -> _SEM:
        """Utility functionality for converting the offset and length of entities from
        Unicode (:obj:`str`) to UTF-16 (``utf-16-le`` encoded :obj:`bytes`).

        Tip:
            Only the offsets and lengths calulated in UTF-16 is acceptable by the Telegram Bot API.
            If they are calculated using the Unicode string (:obj:`str` object), errors may occur
            when the text contains characters that are not in the Basic Multilingual Plane (BMP).
            For more information, see `Unicode <https://en.wikipedia.org/wiki/Unicode>`_ and
            `Plane (Unicode) <https://en.wikipedia.org/wiki/Plane_(Unicode)>`_.

        .. versionadded:: 21.4

        Examples:
            Below is a snippet of code that demonstrates how to use this function to convert
            entities from Unicode to UTF-16 space. The ``unicode_entities`` are calculated in
            Unicode and the `utf_16_entities` are calculated in UTF-16.

            .. code-block:: python

                text = "𠌕 bold 𝄢 italic underlined: 𝛙𝌢𑁍"
                unicode_entities = [
                    MessageEntity(offset=2, length=4, type=MessageEntity.BOLD),
                    MessageEntity(offset=9, length=6, type=MessageEntity.ITALIC),
                    MessageEntity(offset=28, length=3, type=MessageEntity.UNDERLINE),
                ]
                utf_16_entities = MessageEntity.adjust_message_entities_to_utf_16(
                    text, unicode_entities
                )
                await bot.send_message(
                    chat_id=123,
                    text=text,
                    entities=utf_16_entities,
                )
                # utf_16_entities[0]: offset=3, length=4
                # utf_16_entities[1]: offset=11, length=6
                # utf_16_entities[2]: offset=30, length=6

        Args:
            text (:obj:`str`): The text that the entities belong to
            entities (Sequence[:class:`telegram.MessageEntity`]): Sequence of entities
                with offset and length calculated in Unicode

        Returns:
            Sequence[:class:`telegram.MessageEntity`]: Sequence of entities
            with offset and length calculated in UTF-16 encoding
        """
        # get sorted positions
        positions = sorted(itertools.chain(*((x.offset, x.offset + x.length) for x in entities)))
        accumulated_length = 0
        # calculate the length of each slice text[:position] in utf-16 accordingly,
        # store the position translations
        position_translation: dict[int, int] = {}
        for i, position in enumerate(positions):
            last_position = positions[i - 1] if i > 0 else 0
            text_slice = text[last_position:position]
            accumulated_length += len(text_slice.encode(TextEncoding.UTF_16_LE)) // 2
            position_translation[position] = accumulated_length
        # get the final output entities
        out = []
        for entity in entities:
            translated_positions = position_translation[entity.offset]
            translated_length = (
                position_translation[entity.offset + entity.length] - translated_positions
            )
            new_entity = copy.copy(entity)
            with new_entity._unfrozen():
                new_entity.offset = translated_positions
                new_entity.length = translated_length
            out.append(new_entity)
        return out

    @staticmethod
    def shift_entities(by: str | int, entities: _SEM) -> _SEM:
        """Utility functionality for shifting the offset of entities by a given amount.

        Examples:
            Shifting by an integer amount:

            .. code-block:: python

                text = "Hello, world!"
                entities = [
                    MessageEntity(offset=0, length=5, type=MessageEntity.BOLD),
                    MessageEntity(offset=7, length=5, type=MessageEntity.ITALIC),
                ]
                shifted_entities = MessageEntity.shift_entities(1, entities)
                await bot.send_message(
                    chat_id=123,
                    text="!" + text,
                    entities=shifted_entities,
                )

            Shifting using a string:

            .. code-block:: python

                text = "Hello, world!"
                prefix = "𝄢"
                entities = [
                    MessageEntity(offset=0, length=5, type=MessageEntity.BOLD),
                    MessageEntity(offset=7, length=5, type=MessageEntity.ITALIC),
                ]
                shifted_entities = MessageEntity.shift_entities(prefix, entities)
                await bot.send_message(
                    chat_id=123,
                    text=prefix + text,
                    entities=shifted_entities,
                )

        Tip:
            The :paramref:`entities` are *not* modified in place. The function returns a sequence
            of new objects.

        .. versionadded:: 21.5

        Args:
            by (:obj:`str` | :obj:`int`): Either the amount to shift the offset by or
                a string whose length will be used as the amount to shift the offset by. In this
                case, UTF-16 encoding will be used to calculate the length.
            entities (Sequence[:class:`telegram.MessageEntity`]): Sequence of entities

        Returns:
            Sequence[:class:`telegram.MessageEntity`]: Sequence of entities with the offset shifted
        """
        effective_shift = by if isinstance(by, int) else len(by.encode("utf-16-le")) // 2

        out = []
        for entity in entities:
            new_entity = copy.copy(entity)
            with new_entity._unfrozen():
                new_entity.offset += effective_shift
            out.append(new_entity)
        return out

    @classmethod
    def concatenate(
        cls,
        *args: tuple[str, _SEM] | tuple[str, _SEM, bool],
    ) -> tuple[str, _SEM]:
        """Utility functionality for concatenating two text along with their formatting entities.

        Tip:
            This function is useful for prefixing an already formatted text with a new text and its
            formatting entities. In particular, it automatically correctly handles UTF-16 encoding.

        Examples:
            This example shows a callback function that can be used to add a prefix and suffix to
            the message in a :class:`~telegram.ext.CallbackQueryHandler`:

            .. code-block:: python

                async def prefix_message(update: Update, context: ContextTypes.DEFAULT_TYPE):
                    prefix = "𠌕 bold 𝄢 italic underlined: 𝛙𝌢𑁍 | "
                    prefix_entities = [
                        MessageEntity(offset=2, length=4, type=MessageEntity.BOLD),
                        MessageEntity(offset=9, length=6, type=MessageEntity.ITALIC),
                        MessageEntity(offset=28, length=3, type=MessageEntity.UNDERLINE),
                    ]
                    suffix = " | 𠌕 bold 𝄢 italic underlined: 𝛙𝌢𑁍"
                    suffix_entities = [
                        MessageEntity(offset=5, length=4, type=MessageEntity.BOLD),
                        MessageEntity(offset=12, length=6, type=MessageEntity.ITALIC),
                        MessageEntity(offset=31, length=3, type=MessageEntity.UNDERLINE),
                    ]

                    message = update.effective_message
                    first = (prefix, prefix_entities, True)
                    second = (message.text, message.entities)
                    third = (suffix, suffix_entities, True)

                    new_text, new_entities = MessageEntity.concatenate(first, second, third)
                    await update.callback_query.edit_message_text(
                        text=new_text,
                        entities=new_entities,
                    )

        Hint:
            The entities are *not* modified in place. The function returns a
            new sequence of objects.

        .. versionadded:: 21.5

        Args:
            *args (tuple[:obj:`str`, Sequence[:class:`telegram.MessageEntity`]] | \
                tuple[:obj:`str`, Sequence[:class:`telegram.MessageEntity`], :obj:`bool`]):
                Arbitrary number of tuples containing the text and its entities to concatenate.
                If the last element of the tuple is a :obj:`bool`, it is used to determine whether
                to adjust the entities to UTF-16 via
                :meth:`adjust_message_entities_to_utf_16`. UTF-16 adjustment is disabled by
                default.

        Returns:
            tuple[:obj:`str`, Sequence[:class:`telegram.MessageEntity`]]: The concatenated text
            and its entities
        """
        output_text = ""
        output_entities: list[MessageEntity] = []
        for arg in args:
            text, entities = arg[0], arg[1]

            if len(arg) > 2 and arg[2] is True:
                entities = cls.adjust_message_entities_to_utf_16(text, entities)

            output_entities.extend(cls.shift_entities(output_text, entities))
            output_text += text

        return output_text, output_entities

    ALL_TYPES: Final[list[str]] = list(constants.MessageEntityType)
    """list[:obj:`str`]: A list of all available message entity types."""
    BLOCKQUOTE: Final[str] = constants.MessageEntityType.BLOCKQUOTE
    """:const:`telegram.constants.MessageEntityType.BLOCKQUOTE`

    .. versionadded:: 20.8
    """
    BOLD: Final[str] = constants.MessageEntityType.BOLD
    """:const:`telegram.constants.MessageEntityType.BOLD`"""
    BOT_COMMAND: Final[str] = constants.MessageEntityType.BOT_COMMAND
    """:const:`telegram.constants.MessageEntityType.BOT_COMMAND`"""
    CASHTAG: Final[str] = constants.MessageEntityType.CASHTAG
    """:const:`telegram.constants.MessageEntityType.CASHTAG`"""
    CODE: Final[str] = constants.MessageEntityType.CODE
    """:const:`telegram.constants.MessageEntityType.CODE`"""
    CUSTOM_EMOJI: Final[str] = constants.MessageEntityType.CUSTOM_EMOJI
    """:const:`telegram.constants.MessageEntityType.CUSTOM_EMOJI`

    .. versionadded:: 20.0
    """
    DATE_TIME: Final[str] = constants.MessageEntityType.DATE_TIME
    """:const:`telegram.constants.MessageEntityType.DATE_TIME`

    .. versionadded:: 22.7
    """
    EMAIL: Final[str] = constants.MessageEntityType.EMAIL
    """:const:`telegram.constants.MessageEntityType.EMAIL`"""
    EXPANDABLE_BLOCKQUOTE: Final[str] = constants.MessageEntityType.EXPANDABLE_BLOCKQUOTE
    """:const:`telegram.constants.MessageEntityType.EXPANDABLE_BLOCKQUOTE`

    .. versionadded:: 21.3
    """
    HASHTAG: Final[str] = constants.MessageEntityType.HASHTAG
    """:const:`telegram.constants.MessageEntityType.HASHTAG`"""
    ITALIC: Final[str] = constants.MessageEntityType.ITALIC
    """:const:`telegram.constants.MessageEntityType.ITALIC`"""
    MENTION: Final[str] = constants.MessageEntityType.MENTION
    """:const:`telegram.constants.MessageEntityType.MENTION`"""
    PHONE_NUMBER: Final[str] = constants.MessageEntityType.PHONE_NUMBER
    """:const:`telegram.constants.MessageEntityType.PHONE_NUMBER`"""
    PRE: Final[str] = constants.MessageEntityType.PRE
    """:const:`telegram.constants.MessageEntityType.PRE`"""
    SPOILER: Final[str] = constants.MessageEntityType.SPOILER
    """:const:`telegram.constants.MessageEntityType.SPOILER`

    .. versionadded:: 13.10
    """
    STRIKETHROUGH: Final[str] = constants.MessageEntityType.STRIKETHROUGH
    """:const:`telegram.constants.MessageEntityType.STRIKETHROUGH`"""
    TEXT_LINK: Final[str] = constants.MessageEntityType.TEXT_LINK
    """:const:`telegram.constants.MessageEntityType.TEXT_LINK`"""
    TEXT_MENTION: Final[str] = constants.MessageEntityType.TEXT_MENTION
    """:const:`telegram.constants.MessageEntityType.TEXT_MENTION`"""
    UNDERLINE: Final[str] = constants.MessageEntityType.UNDERLINE
    """:const:`telegram.constants.MessageEntityType.UNDERLINE`"""
    URL: Final[str] = constants.MessageEntityType.URL
    """:const:`telegram.constants.MessageEntityType.URL`"""



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
import datetime as dtm
import pickle
from collections import defaultdict

import pytest

from telegram.error import (
    BadRequest,
    ChatMigrated,
    Conflict,
    EndPointNotFound,
    Forbidden,
    InvalidToken,
    NetworkError,
    PassportDecryptionError,
    RetryAfter,
    TelegramError,
    TimedOut,
)
from telegram.ext import InvalidCallbackData
from telegram.warnings import PTBDeprecationWarning
from tests.auxil.slots import mro_slots


class TestErrors:
    def test_telegram_error(self):
        with pytest.raises(TelegramError, match=r"^test message$"):
            raise TelegramError("test message")
        with pytest.raises(TelegramError, match=r"^Test message$"):
            raise TelegramError("Error: test message")
        with pytest.raises(TelegramError, match=r"^Test message$"):
            raise TelegramError("[Error]: test message")
        with pytest.raises(TelegramError, match=r"^Test message$"):
            raise TelegramError("Bad Request: test message")

    def test_unauthorized(self):
        with pytest.raises(Forbidden, match="test message"):
            raise Forbidden("test message")
        with pytest.raises(Forbidden, match=r"^Test message$"):
            raise Forbidden("Error: test message")
        with pytest.raises(Forbidden, match=r"^Test message$"):
            raise Forbidden("[Error]: test message")
        with pytest.raises(Forbidden, match=r"^Test message$"):
            raise Forbidden("Bad Request: test message")

    def test_invalid_token(self):
        with pytest.raises(InvalidToken, match="Invalid token"):
            raise InvalidToken

    def test_network_error(self):
        with pytest.raises(NetworkError, match="test message"):
            raise NetworkError("test message")
        with pytest.raises(NetworkError, match=r"^Test message$"):
            raise NetworkError("Error: test message")
        with pytest.raises(NetworkError, match=r"^Test message$"):
            raise NetworkError("[Error]: test message")
        with pytest.raises(NetworkError, match=r"^Test message$"):
            raise NetworkError("Bad Request: test message")

    def test_bad_request(self):
        with pytest.raises(BadRequest, match="test message"):
            raise BadRequest("test message")
        with pytest.raises(BadRequest, match=r"^Test message$"):
            raise BadRequest("Error: test message")
        with pytest.raises(BadRequest, match=r"^Test message$"):
            raise BadRequest("[Error]: test message")
        with pytest.raises(BadRequest, match=r"^Test message$"):
            raise BadRequest("Bad Request: test message")

    def test_timed_out(self):
        with pytest.raises(TimedOut, match=r"^Timed out$"):
            raise TimedOut

    def test_chat_migrated(self):
        with pytest.raises(ChatMigrated, match="New chat id: 1234") as e:
            raise ChatMigrated(1234)
        assert e.value.new_chat_id == 1234

    @pytest.mark.parametrize("retry_after", [12, dtm.timedelta(seconds=12)])
    def test_retry_after(self, PTB_TIMEDELTA, retry_after):
        if PTB_TIMEDELTA:
            with pytest.raises(RetryAfter, match="Flood control exceeded\\. Retry in 0:00:12"):
                raise (exception := RetryAfter(retry_after))
            assert type(exception.retry_after) is dtm.timedelta
        else:
            with pytest.raises(RetryAfter, match="Flood control exceeded\\. Retry in 12 seconds"):
                raise (exception := RetryAfter(retry_after))
            assert type(exception.retry_after) is int

    def test_retry_after_int_deprecated(self, PTB_TIMEDELTA, recwarn):
        retry_after = RetryAfter(12).retry_after

        if PTB_TIMEDELTA:
            assert len(recwarn) == 0
            assert type(retry_after) is dtm.timedelta
        else:
            assert len(recwarn) == 1
            assert "`retry_after` will be of type `datetime.timedelta`" in str(recwarn[0].message)
            assert recwarn[0].category is PTBDeprecationWarning
            assert type(retry_after) is int

    def test_conflict(self):
        with pytest.raises(Conflict, match="Something something\\."):
            raise Conflict("Something something.")

    @pytest.mark.parametrize(
        ("exception", "attributes"),
        [
            (TelegramError("test message"), ["message"]),
            (Forbidden("test message"), ["message"]),
            (InvalidToken(), ["message"]),
            (NetworkError("test message"), ["message"]),
            (BadRequest("test message"), ["message"]),
            (TimedOut(), ["message"]),
            (ChatMigrated(1234), ["message", "new_chat_id"]),
            (RetryAfter(12), ["message", "retry_after"]),
            (RetryAfter(dtm.timedelta(seconds=12)), ["message", "retry_after"]),
            (Conflict("test message"), ["message"]),
            (PassportDecryptionError("test message"), ["message"]),
            (InvalidCallbackData("test data"), ["callback_data"]),
            (EndPointNotFound("endPoint"), ["message"]),
        ],
    )
    def test_errors_pickling(self, exception, attributes):
        pickled = pickle.dumps(exception)
        unpickled = pickle.loads(pickled)
        assert type(unpickled) is type(exception)
        assert str(unpickled) == str(exception)

        for attribute in attributes:
            assert getattr(unpickled, attribute) == getattr(exception, attribute)

    @pytest.mark.parametrize(
        "inst",
        [
            (TelegramError("test message")),
            (Forbidden("test message")),
            (InvalidToken()),
            (NetworkError("test message")),
            (BadRequest("test message")),
            (TimedOut()),
            (ChatMigrated(1234)),
            (RetryAfter(dtm.timedelta(seconds=12))),
            (Conflict("test message")),
            (PassportDecryptionError("test message")),
            (InvalidCallbackData("test data")),
            (EndPointNotFound("test message")),
        ],
    )
    def test_slot_behaviour(self, inst):
        for attr in inst.__slots__:
            assert getattr(inst, attr, "err") != "err", f"got extra slot '{attr}'"
        assert len(mro_slots(inst)) == len(set(mro_slots(inst))), "duplicate slot"

    def test_coverage(self):
        """
        This test is only here to make sure that new errors will override __reduce__ and set
        __slots__ properly.
        Add the new error class to the below covered_subclasses dict, if it's covered in the above
        test_errors_pickling and test_slots_behavior tests.
        """

        def make_assertion(cls):
            assert set(cls.__subclasses__()) == covered_subclasses[cls]
            for subcls in cls.__subclasses__():
                make_assertion(subcls)

        covered_subclasses = defaultdict(set)
        covered_subclasses.update(
            {
                TelegramError: {
                    Forbidden,
                    InvalidToken,
                    NetworkError,
                    ChatMigrated,
                    RetryAfter,
                    Conflict,
                    PassportDecryptionError,
                    InvalidCallbackData,
                    EndPointNotFound,
                },
                NetworkError: {BadRequest, TimedOut},
            }
        )

        make_assertion(TelegramError)

    def test_string_representations(self, PTB_TIMEDELTA):
        """We just randomly test a few of the subclasses - should suffice"""
        e = TelegramError("This is a message")
        assert repr(e) == "TelegramError('This is a message')"
        assert str(e) == "This is a message"

        e = RetryAfter(dtm.timedelta(seconds=42))
        if PTB_TIMEDELTA:
            assert repr(e) == "RetryAfter('Flood control exceeded. Retry in 0:00:42')"
            assert str(e) == "Flood control exceeded. Retry in 0:00:42"
        else:
            assert repr(e) == "RetryAfter('Flood control exceeded. Retry in 42 seconds')"
            assert str(e) == "Flood control exceeded. Retry in 42 seconds"

        e = BadRequest("This is a message")
        assert repr(e) == "BadRequest('This is a message')"
        assert str(e) == "This is a message"



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
# pylint: disable=missing-module-docstring
from typing import Final, NamedTuple

__all__ = ("__version__", "__version_info__")


class Version(NamedTuple):
    """Copies the behavior of sys.version_info.
    serial is always 0 for stable releases.
    """

    major: int
    minor: int
    micro: int
    releaselevel: str  # Literal['alpha', 'beta', 'candidate', 'final']
    serial: int

    def _rl_shorthand(self) -> str:
        return {
            "alpha": "a",
            "beta": "b",
            "candidate": "rc",
        }[self.releaselevel]

    def __str__(self) -> str:
        version = f"{self.major}.{self.minor}"
        if self.micro != 0:
            version = f"{version}.{self.micro}"
        if self.releaselevel != "final":
            version = f"{version}{self._rl_shorthand()}{self.serial}"

        return version


__version_info__: Final[Version] = Version(
    major=22, minor=7, micro=0, releaselevel="final", serial=0
)
__version__: Final[str] = str(__version_info__)



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

import pytest

from telegram import EncryptedPassportElement, PassportElementError, PassportFile
from tests.auxil.slots import mro_slots


@pytest.fixture(scope="module")
def encrypted_passport_element():
    return EncryptedPassportElement(
        EncryptedPassportElementTestBase.type_,
        "this is a hash",
        data=EncryptedPassportElementTestBase.data,
        phone_number=EncryptedPassportElementTestBase.phone_number,
        email=EncryptedPassportElementTestBase.email,
        files=EncryptedPassportElementTestBase.files,
        front_side=EncryptedPassportElementTestBase.front_side,
        reverse_side=EncryptedPassportElementTestBase.reverse_side,
        selfie=EncryptedPassportElementTestBase.selfie,
    )


class EncryptedPassportElementTestBase:
    type_ = "type"
    hash = "this is a hash"
    data = "data"
    phone_number = "phone_number"
    email = "email"
    files = [PassportFile("file_id", 50, 0, 25)]
    front_side = PassportFile("file_id", 50, 0, 25)
    reverse_side = PassportFile("file_id", 50, 0, 25)
    selfie = PassportFile("file_id", 50, 0, 25)


class TestEncryptedPassportElementWithoutRequest(EncryptedPassportElementTestBase):
    def test_slot_behaviour(self, encrypted_passport_element):
        inst = encrypted_passport_element
        for attr in inst.__slots__:
            assert getattr(inst, attr, "err") != "err", f"got extra slot '{attr}'"
        assert len(mro_slots(inst)) == len(set(mro_slots(inst))), "duplicate slot"

    def test_expected_values(self, encrypted_passport_element):
        assert encrypted_passport_element.type == self.type_
        assert encrypted_passport_element.hash == self.hash
        assert encrypted_passport_element.data == self.data
        assert encrypted_passport_element.phone_number == self.phone_number
        assert encrypted_passport_element.email == self.email
        assert encrypted_passport_element.files == tuple(self.files)
        assert encrypted_passport_element.front_side == self.front_side
        assert encrypted_passport_element.reverse_side == self.reverse_side
        assert encrypted_passport_element.selfie == self.selfie

    def test_to_dict(self, encrypted_passport_element):
        encrypted_passport_element_dict = encrypted_passport_element.to_dict()

        assert isinstance(encrypted_passport_element_dict, dict)
        assert encrypted_passport_element_dict["type"] == encrypted_passport_element.type
        assert encrypted_passport_element_dict["data"] == encrypted_passport_element.data
        assert (
            encrypted_passport_element_dict["phone_number"]
            == encrypted_passport_element.phone_number
        )
        assert encrypted_passport_element_dict["email"] == encrypted_passport_element.email
        assert isinstance(encrypted_passport_element_dict["files"], list)
        assert (
            encrypted_passport_element_dict["front_side"]
            == encrypted_passport_element.front_side.to_dict()
        )
        assert (
            encrypted_passport_element_dict["reverse_side"]
            == encrypted_passport_element.reverse_side.to_dict()
        )
        assert (
            encrypted_passport_element_dict["selfie"]
            == encrypted_passport_element.selfie.to_dict()
        )

    def test_attributes_always_tuple(self):
        element = EncryptedPassportElement(self.type_, self.hash)
        assert element.files == ()
        assert element.translation == ()

    def test_equality(self):
        a = EncryptedPassportElement(self.type_, self.hash, data=self.data)
        b = EncryptedPassportElement(self.type_, self.hash, data=self.data)
        c = EncryptedPassportElement(self.data, "")
        d = PassportElementError("source", "type", "message")

        assert a == b
        assert hash(a) == hash(b)
        assert a is not b

        assert a != c
        assert hash(a) != hash(c)

        assert a != d
        assert hash(a) != hash(d)



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains an object that represents a Telegram Poll."""

import datetime as dtm
from collections.abc import Sequence
from typing import TYPE_CHECKING, Final

from telegram import constants
from telegram._chat import Chat
from telegram._messageentity import MessageEntity
from telegram._telegramobject import TelegramObject
from telegram._user import User
from telegram._utils import enum
from telegram._utils.argumentparsing import (
    de_json_optional,
    de_list_optional,
    parse_sequence_arg,
    to_timedelta,
)
from telegram._utils.datetime import (
    extract_tzinfo_from_defaults,
    from_timestamp,
    get_timedelta_value,
)
from telegram._utils.defaultvalue import DEFAULT_NONE
from telegram._utils.entities import parse_message_entities, parse_message_entity
from telegram._utils.types import JSONDict, ODVInput, TimePeriod

if TYPE_CHECKING:
    from telegram import Bot


class InputPollOption(TelegramObject):
    """
    This object contains information about one answer option in a poll to be sent.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`text` is equal.

    .. versionadded:: 21.2

    Args:
        text (:obj:`str`): Option text,
            :tg-const:`telegram.PollOption.MIN_LENGTH`-:tg-const:`telegram.PollOption.MAX_LENGTH`
            characters.
        text_parse_mode (:obj:`str`, optional): |parse_mode|
            Currently, only custom emoji entities are allowed.
        text_entities (Sequence[:class:`telegram.MessageEntity`], optional): Special entities
            that appear in the option :paramref:`text`. It can be specified instead of
            :paramref:`text_parse_mode`.
            Currently, only custom emoji entities are allowed.
            This list is empty if the text does not contain entities.

    Attributes:
        text (:obj:`str`): Option text,
            :tg-const:`telegram.PollOption.MIN_LENGTH`-:tg-const:`telegram.PollOption.MAX_LENGTH`
            characters.
        text_parse_mode (:obj:`str`): Optional. |parse_mode|
            Currently, only custom emoji entities are allowed.
        text_entities (Sequence[:class:`telegram.MessageEntity`]): Special entities
            that appear in the option :paramref:`text`. It can be specified instead of
            :paramref:`text_parse_mode`.
            Currently, only custom emoji entities are allowed.
            This list is empty if the text does not contain entities.
    """

    __slots__ = ("text", "text_entities", "text_parse_mode")

    def __init__(
        self,
        text: str,
        text_parse_mode: ODVInput[str] = DEFAULT_NONE,
        text_entities: Sequence[MessageEntity] | None = None,
        *,
        api_kwargs: JSONDict | None = None,
    ):
        super().__init__(api_kwargs=api_kwargs)
        self.text: str = text
        self.text_parse_mode: ODVInput[str] = text_parse_mode
        self.text_entities: tuple[MessageEntity, ...] = parse_sequence_arg(text_entities)

        self._id_attrs = (self.text,)

        self._freeze()

    @classmethod
    def de_json(cls, data: JSONDict, bot: "Bot | None" = None) -> "InputPollOption":
        """See :meth:`telegram.TelegramObject.de_json`."""
        data = cls._parse_data(data)

        data["text_entities"] = de_list_optional(data.get("text_entities"), MessageEntity, bot)

        return super().de_json(data=data, bot=bot)


class PollOption(TelegramObject):
    """
    This object contains information about one answer option in a poll.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`text` and :attr:`voter_count` are equal.

    Args:
        text (:obj:`str`): Option text,
            :tg-const:`telegram.PollOption.MIN_LENGTH`-:tg-const:`telegram.PollOption.MAX_LENGTH`
            characters.
        voter_count (:obj:`int`): Number of users that voted for this option.
        text_entities (Sequence[:class:`telegram.MessageEntity`], optional): Special entities
            that appear in the option text. Currently, only custom emoji entities are allowed in
            poll option texts.

            .. versionadded:: 21.2

    Attributes:
        text (:obj:`str`): Option text,
            :tg-const:`telegram.PollOption.MIN_LENGTH`-:tg-const:`telegram.PollOption.MAX_LENGTH`
            characters.
        voter_count (:obj:`int`): Number of users that voted for this option.
        text_entities (tuple[:class:`telegram.MessageEntity`]): Special entities
            that appear in the option text. Currently, only custom emoji entities are allowed in
            poll option texts.
            This list is empty if the question does not contain entities.

            .. versionadded:: 21.2

    """

    __slots__ = ("text", "text_entities", "voter_count")

    def __init__(
        self,
        text: str,
        voter_count: int,
        text_entities: Sequence[MessageEntity] | None = None,
        *,
        api_kwargs: JSONDict | None = None,
    ):
        super().__init__(api_kwargs=api_kwargs)
        self.text: str = text
        self.voter_count: int = voter_count
        self.text_entities: tuple[MessageEntity, ...] = parse_sequence_arg(text_entities)

        self._id_attrs = (self.text, self.voter_count)

        self._freeze()

    @classmethod
    def de_json(cls, data: JSONDict, bot: "Bot | None" = None) -> "PollOption":
        """See :meth:`telegram.TelegramObject.de_json`."""
        data = cls._parse_data(data)

        data["text_entities"] = de_list_optional(data.get("text_entities"), MessageEntity, bot)

        return super().de_json(data=data, bot=bot)

    def parse_entity(self, entity: MessageEntity) -> str:
        """Returns the text in :attr:`text`
        from a given :class:`telegram.MessageEntity` of :attr:`text_entities`.

        Note:
            This method is present because Telegram calculates the offset and length in
            UTF-16 codepoint pairs, which some versions of Python don't handle automatically.
            (That is, you can't just slice ``Message.text`` with the offset and length.)

        .. versionadded:: 21.2

        Args:
            entity (:class:`telegram.MessageEntity`): The entity to extract the text from. It must
                be an entity that belongs to :attr:`text_entities`.

        Returns:
            :obj:`str`: The text of the given entity.
        """
        return parse_message_entity(self.text, entity)

    def parse_entities(self, types: list[str] | None = None) -> dict[MessageEntity, str]:
        """
        Returns a :obj:`dict` that maps :class:`telegram.MessageEntity` to :obj:`str`.
        It contains entities from this polls question filtered by their ``type`` attribute as
        the key, and the text that each entity belongs to as the value of the :obj:`dict`.

        Note:
            This method should always be used instead of the :attr:`text_entities`
            attribute, since it calculates the correct substring from the message text based on
            UTF-16 codepoints. See :attr:`parse_entity` for more info.

        .. versionadded:: 21.2

        Args:
            types (list[:obj:`str`], optional): List of ``MessageEntity`` types as strings. If the
                    ``type`` attribute of an entity is contained in this list, it will be returned.
                    Defaults to :attr:`telegram.MessageEntity.ALL_TYPES`.

        Returns:
            dict[:class:`telegram.MessageEntity`, :obj:`str`]: A dictionary of entities mapped to
            the text that belongs to them, calculated based on UTF-16 codepoints.
        """
        return parse_message_entities(self.text, self.text_entities, types)

    MIN_LENGTH: Final[int] = constants.PollLimit.MIN_OPTION_LENGTH
    """:const:`telegram.constants.PollLimit.MIN_OPTION_LENGTH`

    .. versionadded:: 20.0
    """
    MAX_LENGTH: Final[int] = constants.PollLimit.MAX_OPTION_LENGTH
    """:const:`telegram.constants.PollLimit.MAX_OPTION_LENGTH`

    .. versionadded:: 20.0
    """


class PollAnswer(TelegramObject):
    """
    This object represents an answer of a user in a non-anonymous poll.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`poll_id`, :attr:`user` and :attr:`option_ids` are equal.

    .. versionchanged:: 20.5
        The order of :paramref:`option_ids` and :paramref:`user` is changed in
        20.5 as the latter one became optional.

    .. versionchanged:: 20.6
       Backward compatiblity for changed order of :paramref:`option_ids` and :paramref:`user`
       was removed.

    Args:
        poll_id (:obj:`str`): Unique poll identifier.
        option_ids (Sequence[:obj:`int`]): Identifiers of answer options, chosen by the user. May
            be empty if the user retracted their vote.

            .. versionchanged:: 20.0
                |sequenceclassargs|
        user (:class:`telegram.User`, optional): The user that changed the answer to the poll,
            if the voter isn't anonymous. If the voter is anonymous, this field will contain the
            user :tg-const:`telegram.constants.ChatID.FAKE_CHANNEL` for backwards compatibility.

            .. versionchanged:: 20.5
                :paramref:`user` became optional.
        voter_chat (:class:`telegram.Chat`, optional): The chat that changed the answer to the
            poll, if the voter is anonymous.

            .. versionadded:: 20.5

    Attributes:
        poll_id (:obj:`str`): Unique poll identifier.
        option_ids (tuple[:obj:`int`]): Identifiers of answer options, chosen by the user. May
            be empty if the user retracted their vote.

            .. versionchanged:: 20.0
                |tupleclassattrs|
        user (:class:`telegram.User`): Optional. The user, who changed the answer to the
            poll, if the voter isn't anonymous. If the voter is anonymous, this field will contain
            the user :tg-const:`telegram.constants.ChatID.FAKE_CHANNEL` for backwards compatibility

            .. versionchanged:: 20.5
                :paramref:`user` became optional.
        voter_chat (:class:`telegram.Chat`): Optional. The chat that changed the answer to the
            poll, if the voter is anonymous.

            .. versionadded:: 20.5

    """

    __slots__ = ("option_ids", "poll_id", "user", "voter_chat")

    def __init__(
        self,
        poll_id: str,
        option_ids: Sequence[int],
        user: User | None = None,
        voter_chat: Chat | None = None,
        *,
        api_kwargs: JSONDict | None = None,
    ):
        super().__init__(api_kwargs=api_kwargs)
        self.poll_id: str = poll_id
        self.voter_chat: Chat | None = voter_chat
        self.option_ids: tuple[int, ...] = parse_sequence_arg(option_ids)
        self.user: User | None = user

        self._id_attrs = (
            self.poll_id,
            self.option_ids,
            self.user,
            self.voter_chat,
        )

        self._freeze()

    @classmethod
    def de_json(cls, data: JSONDict, bot: "Bot | None" = None) -> "PollAnswer":
        """See :meth:`telegram.TelegramObject.de_json`."""
        data = cls._parse_data(data)

        data["user"] = de_json_optional(data.get("user"), User, bot)
        data["voter_chat"] = de_json_optional(data.get("voter_chat"), Chat, bot)

        return super().de_json(data=data, bot=bot)


class Poll(TelegramObject):
    """
    This object contains information about a poll.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`id` is equal.

    Examples:
        :any:`Poll Bot <examples.pollbot>`

    Args:
        id (:obj:`str`): Unique poll identifier.
        question (:obj:`str`): Poll question, :tg-const:`telegram.Poll.MIN_QUESTION_LENGTH`-
            :tg-const:`telegram.Poll.MAX_QUESTION_LENGTH` characters.
        options (Sequence[:class:`~telegram.PollOption`]): List of poll options.

            .. versionchanged:: 20.0
                |sequenceclassargs|
        is_closed (:obj:`bool`): :obj:`True`, if the poll is closed.
        is_anonymous (:obj:`bool`): :obj:`True`, if the poll is anonymous.
        type (:obj:`str`): Poll type, currently can be :attr:`REGULAR` or :attr:`QUIZ`.
        allows_multiple_answers (:obj:`bool`): :obj:`True`, if the poll allows multiple answers.
        correct_option_id (:obj:`int`, optional): A zero based identifier of the correct answer
            option. Available only for closed polls in the quiz mode, which were sent
            (not forwarded), by the bot or to a private chat with the bot.
        explanation (:obj:`str`, optional): Text that is shown when a user chooses an incorrect
            answer or taps on the lamp icon in a quiz-style poll,
            0-:tg-const:`telegram.Poll.MAX_EXPLANATION_LENGTH` characters.
        explanation_entities (Sequence[:class:`telegram.MessageEntity`], optional): Special
            entities like usernames, URLs, bot commands, etc. that appear in the
            :attr:`explanation`. This list is empty if the message does not contain explanation
            entities.

            .. versionchanged:: 20.0

               * This attribute is now always a (possibly empty) list and never :obj:`None`.
               * |sequenceclassargs|
        open_period (:obj:`int` | :class:`datetime.timedelta`, optional): Amount of time in seconds
            the poll will be active after creation.

            .. versionchanged:: v22.2
                |time-period-input|
        close_date (:obj:`datetime.datetime`, optional): Point in time (Unix timestamp) when the
            poll will be automatically closed. Converted to :obj:`datetime.datetime`.

            .. versionchanged:: 20.3
                |datetime_localization|
        question_entities (Sequence[:class:`telegram.MessageEntity`], optional): Special entities
            that appear in the :attr:`question`. Currently, only custom emoji entities are allowed
            in poll questions.

            .. versionadded:: 21.2

    Attributes:
        id (:obj:`str`): Unique poll identifier.
        question (:obj:`str`): Poll question, :tg-const:`telegram.Poll.MIN_QUESTION_LENGTH`-
            :tg-const:`telegram.Poll.MAX_QUESTION_LENGTH` characters.
        options (tuple[:class:`~telegram.PollOption`]): List of poll options.

            .. versionchanged:: 20.0
                |tupleclassattrs|
        total_voter_count (:obj:`int`): Total number of users that voted in the poll.
        is_closed (:obj:`bool`): :obj:`True`, if the poll is closed.
        is_anonymous (:obj:`bool`): :obj:`True`, if the poll is anonymous.
        type (:obj:`str`): Poll type, currently can be :attr:`REGULAR` or :attr:`QUIZ`.
        allows_multiple_answers (:obj:`bool`): :obj:`True`, if the poll allows multiple answers.
        correct_option_id (:obj:`int`): Optional. A zero based identifier of the correct answer
            option. Available only for closed polls in the quiz mode, which were sent
            (not forwarded), by the bot or to a private chat with the bot.
        explanation (:obj:`str`): Optional. Text that is shown when a user chooses an incorrect
            answer or taps on the lamp icon in a quiz-style poll,
            0-:tg-const:`telegram.Poll.MAX_EXPLANATION_LENGTH` characters.
        explanation_entities (tuple[:class:`telegram.MessageEntity`]): Special entities
            like usernames, URLs, bot commands, etc. that appear in the :attr:`explanation`.
            This list is empty if the message does not contain explanation entities.

            .. versionchanged:: 20.0
                |tupleclassattrs|

            .. versionchanged:: 20.0
               This attribute is now always a (possibly empty) list and never :obj:`None`.
        open_period (:obj:`int` | :class:`datetime.timedelta`): Optional. Amount of time in seconds
            the poll will be active after creation.

            .. deprecated:: v22.2
                |time-period-int-deprecated|
        close_date (:obj:`datetime.datetime`): Optional. Point in time when the poll will be
            automatically closed.

            .. versionchanged:: 20.3
                |datetime_localization|
        question_entities (tuple[:class:`telegram.MessageEntity`]): Special entities
            that appear in the :attr:`question`. Currently, only custom emoji entities are allowed
            in poll questions.
            This list is empty if the question does not contain entities.

            .. versionadded:: 21.2

    """

    __slots__ = (
        "_open_period",
        "allows_multiple_answers",
        "close_date",
        "correct_option_id",
        "explanation",
        "explanation_entities",
        "id",
        "is_anonymous",
        "is_closed",
        "options",
        "question",
        "question_entities",
        "total_voter_count",
        "type",
    )

    def __init__(
        self,
        id: str,  # pylint: disable=redefined-builtin
        question: str,
        options: Sequence[PollOption],
        total_voter_count: int,
        is_closed: bool,
        is_anonymous: bool,
        type: str,  # pylint: disable=redefined-builtin
        allows_multiple_answers: bool,
        correct_option_id: int | None = None,
        explanation: str | None = None,
        explanation_entities: Sequence[MessageEntity] | None = None,
        open_period: TimePeriod | None = None,
        close_date: dtm.datetime | None = None,
        question_entities: Sequence[MessageEntity] | None = None,
        *,
        api_kwargs: JSONDict | None = None,
    ):
        super().__init__(api_kwargs=api_kwargs)
        self.id: str = id
        self.question: str = question
        self.options: tuple[PollOption, ...] = parse_sequence_arg(options)
        self.total_voter_count: int = total_voter_count
        self.is_closed: bool = is_closed
        self.is_anonymous: bool = is_anonymous
        self.type: str = enum.get_member(constants.PollType, type, type)
        self.allows_multiple_answers: bool = allows_multiple_answers
        self.correct_option_id: int | None = correct_option_id
        self.explanation: str | None = explanation
        self.explanation_entities: tuple[MessageEntity, ...] = parse_sequence_arg(
            explanation_entities
        )
        self._open_period: dtm.timedelta | None = to_timedelta(open_period)
        self.close_date: dtm.datetime | None = close_date
        self.question_entities: tuple[MessageEntity, ...] = parse_sequence_arg(question_entities)

        self._id_attrs = (self.id,)

        self._freeze()

    @property
    def open_period(self) -> int | dtm.timedelta | None:
        return get_timedelta_value(self._open_period, attribute="open_period")

    @classmethod
    def de_json(cls, data: JSONDict, bot: "Bot | None" = None) -> "Poll":
        """See :meth:`telegram.TelegramObject.de_json`."""
        data = cls._parse_data(data)

        # Get the local timezone from the bot if it has defaults
        loc_tzinfo = extract_tzinfo_from_defaults(bot)

        data["options"] = de_list_optional(data.get("options"), PollOption, bot)
        data["explanation_entities"] = de_list_optional(
            data.get("explanation_entities"), MessageEntity, bot
        )
        data["close_date"] = from_timestamp(data.get("close_date"), tzinfo=loc_tzinfo)
        data["question_entities"] = de_list_optional(
            data.get("question_entities"), MessageEntity, bot
        )

        return super().de_json(data=data, bot=bot)

    def parse_explanation_entity(self, entity: MessageEntity) -> str:
        """Returns the text in :attr:`explanation` from a given :class:`telegram.MessageEntity` of
        :attr:`explanation_entities`.

        Note:
            This method is present because Telegram calculates the offset and length in
            UTF-16 codepoint pairs, which some versions of Python don't handle automatically.
            (That is, you can't just slice ``Message.text`` with the offset and length.)

        Args:
            entity (:class:`telegram.MessageEntity`): The entity to extract the text from. It must
                be an entity that belongs to :attr:`explanation_entities`.

        Returns:
            :obj:`str`: The text of the given entity.

        Raises:
            RuntimeError: If the poll has no explanation.

        """
        if not self.explanation:
            raise RuntimeError("This Poll has no 'explanation'.")

        return parse_message_entity(self.explanation, entity)

    def parse_explanation_entities(
        self, types: list[str] | None = None
    ) -> dict[MessageEntity, str]:
        """
        Returns a :obj:`dict` that maps :class:`telegram.MessageEntity` to :obj:`str`.
        It contains entities from this polls explanation filtered by their ``type`` attribute as
        the key, and the text that each entity belongs to as the value of the :obj:`dict`.

        Note:
            This method should always be used instead of the :attr:`explanation_entities`
            attribute, since it calculates the correct substring from the message text based on
            UTF-16 codepoints. See :attr:`parse_explanation_entity` for more info.

        Args:
            types (list[:obj:`str`], optional): List of ``MessageEntity`` types as strings. If the
                    ``type`` attribute of an entity is contained in this list, it will be returned.
                    Defaults to :attr:`telegram.MessageEntity.ALL_TYPES`.

        Returns:
            dict[:class:`telegram.MessageEntity`, :obj:`str`]: A dictionary of entities mapped to
            the text that belongs to them, calculated based on UTF-16 codepoints.

        Raises:
            RuntimeError: If the poll has no explanation.

        """
        if not self.explanation:
            raise RuntimeError("This Poll has no 'explanation'.")

        return parse_message_entities(self.explanation, self.explanation_entities, types)

    def parse_question_entity(self, entity: MessageEntity) -> str:
        """Returns the text in :attr:`question` from a given :class:`telegram.MessageEntity` of
        :attr:`question_entities`.

        .. versionadded:: 21.2

        Note:
            This method is present because Telegram calculates the offset and length in
            UTF-16 codepoint pairs, which some versions of Python don't handle automatically.
            (That is, you can't just slice ``Message.text`` with the offset and length.)

        Args:
            entity (:class:`telegram.MessageEntity`): The entity to extract the text from. It must
                be an entity that belongs to :attr:`question_entities`.

        Returns:
            :obj:`str`: The text of the given entity.
        """
        return parse_message_entity(self.question, entity)

    def parse_question_entities(self, types: list[str] | None = None) -> dict[MessageEntity, str]:
        """
        Returns a :obj:`dict` that maps :class:`telegram.MessageEntity` to :obj:`str`.
        It contains entities from this polls question filtered by their ``type`` attribute as
        the key, and the text that each entity belongs to as the value of the :obj:`dict`.

        .. versionadded:: 21.2

        Note:
            This method should always be used instead of the :attr:`question_entities`
            attribute, since it calculates the correct substring from the message text based on
            UTF-16 codepoints. See :attr:`parse_question_entity` for more info.

        Args:
            types (list[:obj:`str`], optional): List of ``MessageEntity`` types as strings. If the
                    ``type`` attribute of an entity is contained in this list, it will be returned.
                    Defaults to :attr:`telegram.MessageEntity.ALL_TYPES`.

        Returns:
            dict[:class:`telegram.MessageEntity`, :obj:`str`]: A dictionary of entities mapped to
            the text that belongs to them, calculated based on UTF-16 codepoints.

        """
        return parse_message_entities(self.question, self.question_entities, types)

    REGULAR: Final[str] = constants.PollType.REGULAR
    """:const:`telegram.constants.PollType.REGULAR`"""
    QUIZ: Final[str] = constants.PollType.QUIZ
    """:const:`telegram.constants.PollType.QUIZ`"""
    MAX_EXPLANATION_LENGTH: Final[int] = constants.PollLimit.MAX_EXPLANATION_LENGTH
    """:const:`telegram.constants.PollLimit.MAX_EXPLANATION_LENGTH`

    .. versionadded:: 20.0
    """
    MAX_EXPLANATION_LINE_FEEDS: Final[int] = constants.PollLimit.MAX_EXPLANATION_LINE_FEEDS
    """:const:`telegram.constants.PollLimit.MAX_EXPLANATION_LINE_FEEDS`

    .. versionadded:: 20.0
    """
    MIN_OPEN_PERIOD: Final[int] = constants.PollLimit.MIN_OPEN_PERIOD
    """:const:`telegram.constants.PollLimit.MIN_OPEN_PERIOD`

    .. versionadded:: 20.0
    """
    MAX_OPEN_PERIOD: Final[int] = constants.PollLimit.MAX_OPEN_PERIOD
    """:const:`telegram.constants.PollLimit.MAX_OPEN_PERIOD`

    .. versionadded:: 20.0
    """
    MIN_QUESTION_LENGTH: Final[int] = constants.PollLimit.MIN_QUESTION_LENGTH
    """:const:`telegram.constants.PollLimit.MIN_QUESTION_LENGTH`

    .. versionadded:: 20.0
    """
    MAX_QUESTION_LENGTH: Final[int] = constants.PollLimit.MAX_QUESTION_LENGTH
    """:const:`telegram.constants.PollLimit.MAX_QUESTION_LENGTH`

    .. versionadded:: 20.0
    """
    MIN_OPTION_LENGTH: Final[int] = constants.PollLimit.MIN_OPTION_LENGTH
    """:const:`telegram.constants.PollLimit.MIN_OPTION_LENGTH`

    .. versionadded:: 20.0
    """
    MAX_OPTION_LENGTH: Final[int] = constants.PollLimit.MAX_OPTION_LENGTH
    """:const:`telegram.constants.PollLimit.MAX_OPTION_LENGTH`

    .. versionadded:: 20.0
    """
    MIN_OPTION_NUMBER: Final[int] = constants.PollLimit.MIN_OPTION_NUMBER
    """:const:`telegram.constants.PollLimit.MIN_OPTION_NUMBER`

    .. versionadded:: 20.0
    """
    MAX_OPTION_NUMBER: Final[int] = constants.PollLimit.MAX_OPTION_NUMBER
    """:const:`telegram.constants.PollLimit.MAX_OPTION_NUMBER`

    .. versionadded:: 20.0
    """



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].
import inspect
from typing import TYPE_CHECKING

import pytest

import telegram
from tests.auxil.envvars import RUN_TEST_OFFICIAL
from tests.test_official.arg_type_checker import (
    check_defaults_type,
    check_param_type,
    check_required_param,
)
from tests.test_official.exceptions import (
    GLOBALLY_IGNORED_PARAMETERS,
    backwards_compat_kwargs,
    ptb_extra_params,
    ptb_ignored_params,
)
from tests.test_official.scraper import Scraper, TelegramClass, TelegramMethod

if TYPE_CHECKING:
    from types import FunctionType

# Will skip all tests in this file if the env var is False
pytestmark = pytest.mark.skipif(not RUN_TEST_OFFICIAL, reason="test_official is not enabled")

methods, method_ids, classes, class_ids = [], [], [], []  # not needed (just for completeness)

if RUN_TEST_OFFICIAL:
    scraper = Scraper()
    methods, method_ids = scraper.collect_methods()
    classes, class_ids = scraper.collect_classes()


@pytest.mark.parametrize("tg_method", argvalues=methods, ids=method_ids)
def test_check_method(tg_method: TelegramMethod) -> None:
    """This function checks for the following things compared to the official API docs:

    - Method existence
    - Parameter existence
    - Parameter requirement correctness
    - Parameter type annotation existence
    - Parameter type annotation correctness
    - Parameter default value correctness
    - No unexpected parameters
    - Extra parameters should be keyword only
    """
    ptb_method: FunctionType | None = getattr(telegram.Bot, tg_method.method_name, None)
    assert ptb_method, f"Method {tg_method.method_name} not found in telegram.Bot"

    # Check arguments based on source
    sig = inspect.signature(ptb_method, follow_wrapped=True)
    checked = []

    for tg_parameter in tg_method.method_parameters:
        # Check if parameter is present in our method
        ptb_param = sig.parameters.get(tg_parameter.param_name)
        assert ptb_param is not None, (
            f"Parameter {tg_parameter.param_name} not found in {ptb_method.__name__}"
        )

        # Now check if the parameter is required or not
        assert check_required_param(tg_parameter, ptb_param, ptb_method.__name__), (
            f"Param {ptb_param.name!r} of {ptb_method.__name__!r} requirement mismatch"
        )

        # Check if type annotation is present
        assert ptb_param.annotation is not inspect.Parameter.empty, (
            f"Param {ptb_param.name!r} of {ptb_method.__name__!r} should have a type annotation!"
        )
        # Check if type annotation is correct
        correct_type_hint, expected_type_hint = check_param_type(
            ptb_param,
            tg_parameter,
            ptb_method,
        )
        assert correct_type_hint, (
            f"Type hint of param {ptb_param.name!r} of {ptb_method.__name__!r} should be "
            f"{expected_type_hint!r} or something else!"
        )

        # Now we will check that we don't pass default values if the parameter is not required.
        if ptb_param.default is not inspect.Parameter.empty:  # If there is a default argument...
            default_arg_none = check_defaults_type(ptb_param)  # check if it's None
            assert default_arg_none, (
                f"Param {ptb_param.name!r} of {ptb_method.__name__!r} should be `None`"
            )
        checked.append(tg_parameter.param_name)

    expected_additional_args = GLOBALLY_IGNORED_PARAMETERS.copy()
    expected_additional_args |= ptb_extra_params(tg_method.method_name)
    expected_additional_args |= backwards_compat_kwargs(tg_method.method_name)

    unexpected_args = (sig.parameters.keys() ^ checked) - expected_additional_args
    assert unexpected_args == set(), (
        f"In {ptb_method.__qualname__}, unexpected args were found: {unexpected_args}."
    )

    kw_or_positional_args = [
        p.name for p in sig.parameters.values() if p.kind != inspect.Parameter.KEYWORD_ONLY
    ]
    non_kw_only_args = set(kw_or_positional_args).difference(checked).difference(["self"])
    non_kw_only_args -= backwards_compat_kwargs(tg_method.method_name)
    assert non_kw_only_args == set(), (
        f"In {ptb_method.__qualname__}, extra args should be keyword only (compared to "
        f"{tg_method.method_name} in API)"
    )


@pytest.mark.parametrize("tg_class", argvalues=classes, ids=class_ids)
def test_check_object(tg_class: TelegramClass) -> None:
    """This function checks for the following things compared to the official API docs:

    - Class existence
    - Parameter existence
    - Parameter requirement correctness
    - Parameter type annotation existence
    - Parameter type annotation correctness
    - Parameter default value correctness
    - No unexpected parameters
    """
    obj = getattr(telegram, tg_class.class_name)

    # Check arguments based on source. Makes sure to only check __init__'s signature & nothing else
    sig = inspect.signature(obj.__init__, follow_wrapped=True)

    checked = set()
    fields_removed_by_ptb = ptb_ignored_params(tg_class.class_name)

    for tg_parameter in tg_class.class_parameters:
        field: str = tg_parameter.param_name

        if field in fields_removed_by_ptb:
            continue

        if field == "from":
            field = "from_user"

        ptb_param = sig.parameters.get(field)
        assert ptb_param is not None, f"Attribute {field} not found in {obj.__name__}"

        # Now check if the parameter is required or not
        assert check_required_param(tg_parameter, ptb_param, obj.__name__), (
            f"Param {ptb_param.name!r} of {obj.__name__!r} requirement mismatch"
        )

        # Check if type annotation is present
        assert ptb_param.annotation is not inspect.Parameter.empty, (
            f"Param {ptb_param.name!r} of {obj.__name__!r} should have a type annotation"
        )

        # Check if type annotation is correct
        correct_type_hint, expected_type_hint = check_param_type(ptb_param, tg_parameter, obj)
        assert correct_type_hint, (
            f"Type hint of param {ptb_param.name!r} of {obj.__name__!r} should be "
            f"{expected_type_hint!r} or something else!"
        )

        # Now we will check that we don't pass default values if the parameter is not required.
        if ptb_param.default is not inspect.Parameter.empty:  # If there is a default argument...
            default_arg_none = check_defaults_type(ptb_param)  # check if its None
            assert default_arg_none, (
                f"Param {ptb_param.name!r} of {obj.__name__!r} should be `None`"
            )

        checked.add(field)

    expected_additional_args = GLOBALLY_IGNORED_PARAMETERS.copy()
    expected_additional_args |= ptb_extra_params(tg_class.class_name)
    expected_additional_args |= backwards_compat_kwargs(tg_class.class_name)

    unexpected_args = (sig.parameters.keys() ^ checked) - expected_additional_args
    assert unexpected_args == set(), (
        f"In {tg_class.class_name}, unexpected args were found: {unexpected_args}."
    )



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains an object that represents a Telegram Update."""

from typing import TYPE_CHECKING, Final

from telegram import constants
from telegram._business import BusinessConnection, BusinessMessagesDeleted
from telegram._callbackquery import CallbackQuery
from telegram._chatboost import ChatBoostRemoved, ChatBoostUpdated
from telegram._chatjoinrequest import ChatJoinRequest
from telegram._chatmemberupdated import ChatMemberUpdated
from telegram._choseninlineresult import ChosenInlineResult
from telegram._inline.inlinequery import InlineQuery
from telegram._message import Message
from telegram._messagereactionupdated import MessageReactionCountUpdated, MessageReactionUpdated
from telegram._paidmedia import PaidMediaPurchased
from telegram._payment.precheckoutquery import PreCheckoutQuery
from telegram._payment.shippingquery import ShippingQuery
from telegram._poll import Poll, PollAnswer
from telegram._telegramobject import TelegramObject
from telegram._utils.argumentparsing import de_json_optional
from telegram._utils.types import JSONDict
from telegram._utils.warnings import warn

if TYPE_CHECKING:
    from telegram import Bot, Chat, User


class Update(TelegramObject):
    """This object represents an incoming update.

    Objects of this class are comparable in terms of equality. Two objects of this class are
    considered equal, if their :attr:`update_id` is equal.

    Note:
        At most one of the optional parameters can be present in any given update.

    .. seealso:: :wiki:`Your First Bot <Extensions---Your-first-Bot>`

    Args:
        update_id (:obj:`int`): The update's unique identifier. Update identifiers start from a
            certain positive number and increase sequentially. This ID becomes especially handy if
            you're using Webhooks, since it allows you to ignore repeated updates or to restore the
            correct update sequence, should they get out of order. If there are no new updates for
            at least a week, then identifier of the next update will be chosen randomly instead of
            sequentially.
        message (:class:`telegram.Message`, optional): New incoming message of any kind - text,
            photo, sticker, etc.
        edited_message (:class:`telegram.Message`, optional): New version of a message that is
            known to the bot and was edited. This update may at times be triggered by changes to
            message fields that are either unavailable or not actively used by your bot.
        channel_post (:class:`telegram.Message`, optional): New incoming channel post of any kind
            - text, photo, sticker, etc.
        edited_channel_post (:class:`telegram.Message`, optional): New version of a channel post
            that is known to the bot and was edited. This update may at times be triggered by
            changes to message fields that are either unavailable or not actively used by your bot.
        inline_query (:class:`telegram.InlineQuery`, optional): New incoming inline query.
        chosen_inline_result (:class:`telegram.ChosenInlineResult`, optional): The result of an
            inline query that was chosen by a user and sent to their chat partner.
        callback_query (:class:`telegram.CallbackQuery`, optional): New incoming callback query.
        shipping_query (:class:`telegram.ShippingQuery`, optional): New incoming shipping query.
            Only for invoices with flexible price.
        pre_checkout_query (:class:`telegram.PreCheckoutQuery`, optional): New incoming
            pre-checkout query. Contains full information about checkout.
        poll (:class:`telegram.Poll`, optional): New poll state. Bots receive only updates about
            manually stopped polls and polls, which are sent by the bot.
        poll_answer (:class:`telegram.PollAnswer`, optional): A user changed their answer
            in a non-anonymous poll. Bots receive new votes only in polls that were sent
            by the bot itself.
        my_chat_member (:class:`telegram.ChatMemberUpdated`, optional): The bot's chat member
            status was updated in a chat. For private chats, this update is received only when the
            bot is blocked or unblocked by the user.

            .. versionadded:: 13.4
        chat_member (:class:`telegram.ChatMemberUpdated`, optional): A chat member's status was
            updated in a chat. The bot must be an administrator in the chat and must explicitly
            specify :attr:`CHAT_MEMBER` in the list of
            :paramref:`telegram.ext.Application.run_polling.allowed_updates` to receive these
            updates (see :meth:`telegram.Bot.get_updates`, :meth:`telegram.Bot.set_webhook`,
            :meth:`telegram.ext.Application.run_polling` and
            :meth:`telegram.ext.Application.run_webhook`).

            .. versionadded:: 13.4
        chat_join_request (:class:`telegram.ChatJoinRequest`, optional): A request to join the
            chat has been sent. The bot must have the
            :attr:`telegram.ChatPermissions.can_invite_users` administrator right in the chat to
            receive these updates.

            .. versionadded:: 13.8

        chat_boost (:class:`telegram.ChatBoostUpdated`, optional): A chat boost was added or
            changed. The bot must be an administrator in the chat to receive these updates.

            .. versionadded:: 20.8

        removed_chat_boost (:class:`telegram.ChatBoostRemoved`, optional): A boost was removed from
            a chat. The bot must be an administrator in the chat to receive these updates.

            .. versionadded:: 20.8

        message_reaction (:class:`telegram.MessageReactionUpdated`, optional): A reaction to a
            message was changed by a user. The bot must be an administrator in the chat and must
            explicitly specify :attr:`MESSAGE_REACTION` in the list of
            :paramref:`telegram.ext.Application.run_polling.allowed_updates` to receive these
            updates (see :meth:`telegram.Bot.get_updates`, :meth:`telegram.Bot.set_webhook`,
            :meth:`telegram.ext.Application.run_polling` and
            :meth:`telegram.ext.Application.run_webhook`). The update isn't received for reactions
            set by bots.

            .. versionadded:: 20.8

        message_reaction_count (:class:`telegram.MessageReactionCountUpdated`, optional): Reactions
            to a message with anonymous reactions were changed. The bot must be an administrator in
            the chat and must explicitly specify :attr:`MESSAGE_REACTION_COUNT` in the list of
            :paramref:`telegram.ext.Application.run_polling.allowed_updates` to receive these
            updates (see :meth:`telegram.Bot.get_updates`, :meth:`telegram.Bot.set_webhook`,
            :meth:`telegram.ext.Application.run_polling` and
            :meth:`telegram.ext.Application.run_webhook`). The updates are grouped and can be sent
            with delay up to a few minutes.

            .. versionadded:: 20.8

        business_connection (:class:`telegram.BusinessConnection`, optional): The bot was connected
            to or disconnected from a business account, or a user edited an existing connection
            with the bot.

            .. versionadded:: 21.1

        business_message (:class:`telegram.Message`, optional): New message from a connected
            business account.

            .. versionadded:: 21.1

        edited_business_message (:class:`telegram.Message`, optional): New version of a message
            from a connected business account.

            .. versionadded:: 21.1

        deleted_business_messages (:class:`telegram.BusinessMessagesDeleted`, optional): Messages
            were deleted from a connected business account.

            .. versionadded:: 21.1

        purchased_paid_media (:class:`telegram.PaidMediaPurchased`, optional): A user purchased
            paid media with a non-empty payload sent by the bot in a non-channel chat.

            .. versionadded:: 21.6


    Attributes:
        update_id (:obj:`int`): The update's unique identifier. Update identifiers start from a
            certain positive number and increase sequentially. This ID becomes especially handy if
            you're using Webhooks, since it allows you to ignore repeated updates or to restore the
            correct update sequence, should they get out of order. If there are no new updates for
            at least a week, then identifier of the next update will be chosen randomly instead of
            sequentially.
        message (:class:`telegram.Message`): Optional. New incoming message of any kind - text,
            photo, sticker, etc.
        edited_message (:class:`telegram.Message`): Optional. New version of a message that is
            known to the bot and was edited. This update may at times be triggered by changes to
            message fields that are either unavailable or not actively used by your bot.
        channel_post (:class:`telegram.Message`): Optional. New incoming channel post of any kind
            - text, photo, sticker, etc.
        edited_channel_post (:class:`telegram.Message`): Optional. New version of a channel post
            that is known to the bot and was edited. This update may at times be triggered by
            changes to message fields that are either unavailable or not actively used by your bot.
        inline_query (:class:`telegram.InlineQuery`): Optional. New incoming inline query.
        chosen_inline_result (:class:`telegram.ChosenInlineResult`): Optional. The result of an
            inline query that was chosen by a user and sent to their chat partner.
        callback_query (:class:`telegram.CallbackQuery`): Optional. New incoming callback query.

            Examples:
                :any:`Arbitrary Callback Data Bot <examples.arbitrarycallbackdatabot>`
        shipping_query (:class:`telegram.ShippingQuery`): Optional. New incoming shipping query.
            Only for invoices with flexible price.
        pre_checkout_query (:class:`telegram.PreCheckoutQuery`): Optional. New incoming
            pre-checkout query. Contains full information about checkout.
        poll (:class:`telegram.Poll`): Optional. New poll state. Bots receive only updates about
            manually stopped polls and polls, which are sent by the bot.
        poll_answer (:class:`telegram.PollAnswer`): Optional. A user changed their answer
            in a non-anonymous poll. Bots receive new votes only in polls that were sent
            by the bot itself.
        my_chat_member (:class:`telegram.ChatMemberUpdated`): Optional. The bot's chat member
            status was updated in a chat. For private chats, this update is received only when the
            bot is blocked or unblocked by the user.

            .. versionadded:: 13.4
        chat_member (:class:`telegram.ChatMemberUpdated`): Optional. A chat member's status was
            updated in a chat. The bot must be an administrator in the chat and must explicitly
            specify :attr:`CHAT_MEMBER` in the list of
            :paramref:`telegram.ext.Application.run_polling.allowed_updates` to receive these
            updates (see :meth:`telegram.Bot.get_updates`, :meth:`telegram.Bot.set_webhook`,
            :meth:`telegram.ext.Application.run_polling` and
            :meth:`telegram.ext.Application.run_webhook`).

            .. versionadded:: 13.4
        chat_join_request (:class:`telegram.ChatJoinRequest`): Optional. A request to join the
            chat has been sent. The bot must have the
            :attr:`telegram.ChatPermissions.can_invite_users` administrator right in the chat to
            receive these updates.

            .. versionadded:: 13.8

        chat_boost (:class:`telegram.ChatBoostUpdated`): Optional. A chat boost was added or
            changed. The bot must be an administrator in the chat to receive these updates.

            .. versionadded:: 20.8

        removed_chat_boost (:class:`telegram.ChatBoostRemoved`): Optional. A boost was removed from
            a chat. The bot must be an administrator in the chat to receive these updates.

            .. versionadded:: 20.8

        message_reaction (:class:`telegram.MessageReactionUpdated`): Optional. A reaction to a
            message was changed by a user. The bot must be an administrator in the chat and must
            explicitly specify :attr:`MESSAGE_REACTION` in the list of
            :paramref:`telegram.ext.Application.run_polling.allowed_updates` to receive these
            updates (see :meth:`telegram.Bot.get_updates`, :meth:`telegram.Bot.set_webhook`,
            :meth:`telegram.ext.Application.run_polling` and
            :meth:`telegram.ext.Application.run_webhook`). The update isn't received for reactions
            set by bots.

            .. versionadded:: 20.8

        message_reaction_count (:class:`telegram.MessageReactionCountUpdated`): Optional. Reactions
            to a message with anonymous reactions were changed. The bot must be an administrator in
            the chat and must explicitly specify :attr:`MESSAGE_REACTION_COUNT` in the list of
            :paramref:`telegram.ext.Application.run_polling.allowed_updates` to receive these
            updates (see :meth:`telegram.Bot.get_updates`, :meth:`telegram.Bot.set_webhook`,
            :meth:`telegram.ext.Application.run_polling` and
            :meth:`telegram.ext.Application.run_webhook`). The updates are grouped and can be sent
            with delay up to a few minutes.

            .. versionadded:: 20.8

        business_connection (:class:`telegram.BusinessConnection`): Optional. The bot was connected
            to or disconnected from a business account, or a user edited an existing connection
            with the bot.

            .. versionadded:: 21.1

        business_message (:class:`telegram.Message`): Optional. New message from a connected
            business account.

            .. versionadded:: 21.1

        edited_business_message (:class:`telegram.Message`): Optional. New version of a message
            from a connected business account.

            .. versionadded:: 21.1

        deleted_business_messages (:class:`telegram.BusinessMessagesDeleted`): Optional. Messages
            were deleted from a connected business account.

            .. versionadded:: 21.1

        purchased_paid_media (:class:`telegram.PaidMediaPurchased`): Optional. A user purchased
            paid media with a non-empty payload sent by the bot in a non-channel chat.

            .. versionadded:: 21.6
    """

    __slots__ = (
        "_effective_chat",
        "_effective_message",
        "_effective_sender",
        "_effective_user",
        "business_connection",
        "business_message",
        "callback_query",
        "channel_post",
        "chat_boost",
        "chat_join_request",
        "chat_member",
        "chosen_inline_result",
        "deleted_business_messages",
        "edited_business_message",
        "edited_channel_post",
        "edited_message",
        "inline_query",
        "message",
        "message_reaction",
        "message_reaction_count",
        "my_chat_member",
        "poll",
        "poll_answer",
        "pre_checkout_query",
        "purchased_paid_media",
        "removed_chat_boost",
        "shipping_query",
        "update_id",
    )

    MESSAGE: Final[str] = constants.UpdateType.MESSAGE
    """:const:`telegram.constants.UpdateType.MESSAGE`

    .. versionadded:: 13.5"""
    EDITED_MESSAGE: Final[str] = constants.UpdateType.EDITED_MESSAGE
    """:const:`telegram.constants.UpdateType.EDITED_MESSAGE`

    .. versionadded:: 13.5"""
    CHANNEL_POST: Final[str] = constants.UpdateType.CHANNEL_POST
    """:const:`telegram.constants.UpdateType.CHANNEL_POST`

    .. versionadded:: 13.5"""
    EDITED_CHANNEL_POST: Final[str] = constants.UpdateType.EDITED_CHANNEL_POST
    """:const:`telegram.constants.UpdateType.EDITED_CHANNEL_POST`

    .. versionadded:: 13.5"""
    INLINE_QUERY: Final[str] = constants.UpdateType.INLINE_QUERY
    """:const:`telegram.constants.UpdateType.INLINE_QUERY`

    .. versionadded:: 13.5"""
    CHOSEN_INLINE_RESULT: Final[str] = constants.UpdateType.CHOSEN_INLINE_RESULT
    """:const:`telegram.constants.UpdateType.CHOSEN_INLINE_RESULT`

    .. versionadded:: 13.5"""
    CALLBACK_QUERY: Final[str] = constants.UpdateType.CALLBACK_QUERY
    """:const:`telegram.constants.UpdateType.CALLBACK_QUERY`

    .. versionadded:: 13.5"""
    SHIPPING_QUERY: Final[str] = constants.UpdateType.SHIPPING_QUERY
    """:const:`telegram.constants.UpdateType.SHIPPING_QUERY`

    .. versionadded:: 13.5"""
    PRE_CHECKOUT_QUERY: Final[str] = constants.UpdateType.PRE_CHECKOUT_QUERY
    """:const:`telegram.constants.UpdateType.PRE_CHECKOUT_QUERY`

    .. versionadded:: 13.5"""
    POLL: Final[str] = constants.UpdateType.POLL
    """:const:`telegram.constants.UpdateType.POLL`

    .. versionadded:: 13.5"""
    POLL_ANSWER: Final[str] = constants.UpdateType.POLL_ANSWER
    """:const:`telegram.constants.UpdateType.POLL_ANSWER`

    .. versionadded:: 13.5"""
    MY_CHAT_MEMBER: Final[str] = constants.UpdateType.MY_CHAT_MEMBER
    """:const:`telegram.constants.UpdateType.MY_CHAT_MEMBER`

    .. versionadded:: 13.5"""
    CHAT_MEMBER: Final[str] = constants.UpdateType.CHAT_MEMBER
    """:const:`telegram.constants.UpdateType.CHAT_MEMBER`

    .. versionadded:: 13.5"""
    CHAT_JOIN_REQUEST: Final[str] = constants.UpdateType.CHAT_JOIN_REQUEST
    """:const:`telegram.constants.UpdateType.CHAT_JOIN_REQUEST`

    .. versionadded:: 13.8"""
    CHAT_BOOST: Final[str] = constants.UpdateType.CHAT_BOOST
    """:const:`telegram.constants.UpdateType.CHAT_BOOST`

    .. versionadded:: 20.8"""
    REMOVED_CHAT_BOOST: Final[str] = constants.UpdateType.REMOVED_CHAT_BOOST
    """:const:`telegram.constants.UpdateType.REMOVED_CHAT_BOOST`

    .. versionadded:: 20.8"""
    MESSAGE_REACTION: Final[str] = constants.UpdateType.MESSAGE_REACTION
    """:const:`telegram.constants.UpdateType.MESSAGE_REACTION`

    .. versionadded:: 20.8"""
    MESSAGE_REACTION_COUNT: Final[str] = constants.UpdateType.MESSAGE_REACTION_COUNT
    """:const:`telegram.constants.UpdateType.MESSAGE_REACTION_COUNT`

    .. versionadded:: 20.8"""
    BUSINESS_CONNECTION: Final[str] = constants.UpdateType.BUSINESS_CONNECTION
    """:const:`telegram.constants.UpdateType.BUSINESS_CONNECTION`

    .. versionadded:: 21.1"""
    BUSINESS_MESSAGE: Final[str] = constants.UpdateType.BUSINESS_MESSAGE
    """:const:`telegram.constants.UpdateType.BUSINESS_MESSAGE`

    .. versionadded:: 21.1"""
    EDITED_BUSINESS_MESSAGE: Final[str] = constants.UpdateType.EDITED_BUSINESS_MESSAGE
    """:const:`telegram.constants.UpdateType.EDITED_BUSINESS_MESSAGE`

    .. versionadded:: 21.1"""
    DELETED_BUSINESS_MESSAGES: Final[str] = constants.UpdateType.DELETED_BUSINESS_MESSAGES
    """:const:`telegram.constants.UpdateType.DELETED_BUSINESS_MESSAGES`

    .. versionadded:: 21.1"""

    PURCHASED_PAID_MEDIA: Final[str] = constants.UpdateType.PURCHASED_PAID_MEDIA
    """:const:`telegram.constants.UpdateType.PURCHASED_PAID_MEDIA`

    .. versionadded:: 21.6
    """

    ALL_TYPES: Final[list[str]] = list(constants.UpdateType)
    """list[:obj:`str`]: A list of all available update types.

    .. versionadded:: 13.5"""

    def __init__(
        self,
        update_id: int,
        message: Message | None = None,
        edited_message: Message | None = None,
        channel_post: Message | None = None,
        edited_channel_post: Message | None = None,
        inline_query: InlineQuery | None = None,
        chosen_inline_result: ChosenInlineResult | None = None,
        callback_query: CallbackQuery | None = None,
        shipping_query: ShippingQuery | None = None,
        pre_checkout_query: PreCheckoutQuery | None = None,
        poll: Poll | None = None,
        poll_answer: PollAnswer | None = None,
        my_chat_member: ChatMemberUpdated | None = None,
        chat_member: ChatMemberUpdated | None = None,
        chat_join_request: ChatJoinRequest | None = None,
        chat_boost: ChatBoostUpdated | None = None,
        removed_chat_boost: ChatBoostRemoved | None = None,
        message_reaction: MessageReactionUpdated | None = None,
        message_reaction_count: MessageReactionCountUpdated | None = None,
        business_connection: BusinessConnection | None = None,
        business_message: Message | None = None,
        edited_business_message: Message | None = None,
        deleted_business_messages: BusinessMessagesDeleted | None = None,
        purchased_paid_media: PaidMediaPurchased | None = None,
        *,
        api_kwargs: JSONDict | None = None,
    ):
        super().__init__(api_kwargs=api_kwargs)
        # Required
        self.update_id: int = update_id
        # Optionals
        self.message: Message | None = message
        self.edited_message: Message | None = edited_message
        self.inline_query: InlineQuery | None = inline_query
        self.chosen_inline_result: ChosenInlineResult | None = chosen_inline_result
        self.callback_query: CallbackQuery | None = callback_query
        self.shipping_query: ShippingQuery | None = shipping_query
        self.pre_checkout_query: PreCheckoutQuery | None = pre_checkout_query
        self.channel_post: Message | None = channel_post
        self.edited_channel_post: Message | None = edited_channel_post
        self.poll: Poll | None = poll
        self.poll_answer: PollAnswer | None = poll_answer
        self.my_chat_member: ChatMemberUpdated | None = my_chat_member
        self.chat_member: ChatMemberUpdated | None = chat_member
        self.chat_join_request: ChatJoinRequest | None = chat_join_request
        self.chat_boost: ChatBoostUpdated | None = chat_boost
        self.removed_chat_boost: ChatBoostRemoved | None = removed_chat_boost
        self.message_reaction: MessageReactionUpdated | None = message_reaction
        self.message_reaction_count: MessageReactionCountUpdated | None = message_reaction_count
        self.business_connection: BusinessConnection | None = business_connection
        self.business_message: Message | None = business_message
        self.edited_business_message: Message | None = edited_business_message
        self.deleted_business_messages: BusinessMessagesDeleted | None = deleted_business_messages
        self.purchased_paid_media: PaidMediaPurchased | None = purchased_paid_media

        self._effective_user: User | None = None
        self._effective_sender: User | Chat | None = None
        self._effective_chat: Chat | None = None
        self._effective_message: Message | None = None

        self._id_attrs = (self.update_id,)

        self._freeze()

    @property
    def effective_user(self) -> "User | None":
        """
        :class:`telegram.User`: The user that sent this update, no matter what kind of update this
        is. If no user is associated with this update, this gives :obj:`None`. This is the case
        if any of

        * :attr:`channel_post`
        * :attr:`edited_channel_post`
        * :attr:`poll`
        * :attr:`chat_boost`
        * :attr:`removed_chat_boost`
        * :attr:`message_reaction_count`
        * :attr:`deleted_business_messages`

        is present.

        .. versionchanged:: 21.1
            This property now also considers :attr:`business_connection`, :attr:`business_message`
            and :attr:`edited_business_message`.

        .. versionchanged:: 21.6
            This property now also considers :attr:`purchased_paid_media`.

        Example:
            * If :attr:`message` is present, this will give
              :attr:`telegram.Message.from_user`.
            * If :attr:`poll_answer` is present, this will give :attr:`telegram.PollAnswer.user`.

        """
        if self._effective_user:
            return self._effective_user

        user = None

        if self.message:
            user = self.message.from_user

        elif self.edited_message:
            user = self.edited_message.from_user

        elif self.inline_query:
            user = self.inline_query.from_user

        elif self.chosen_inline_result:
            user = self.chosen_inline_result.from_user

        elif self.callback_query:
            user = self.callback_query.from_user

        elif self.shipping_query:
            user = self.shipping_query.from_user

        elif self.pre_checkout_query:
            user = self.pre_checkout_query.from_user

        elif self.poll_answer:
            user = self.poll_answer.user

        elif self.my_chat_member:
            user = self.my_chat_member.from_user

        elif self.chat_member:
            user = self.chat_member.from_user

        elif self.chat_join_request:
            user = self.chat_join_request.from_user

        elif self.message_reaction:
            user = self.message_reaction.user

        elif self.business_message:
            user = self.business_message.from_user

        elif self.edited_business_message:
            user = self.edited_business_message.from_user

        elif self.business_connection:
            user = self.business_connection.user

        elif self.purchased_paid_media:
            user = self.purchased_paid_media.from_user

        self._effective_user = user
        return user

    @property
    def effective_sender(self) -> "User | Chat | None":
        """
        :class:`telegram.User` or :class:`telegram.Chat`: The user or chat that sent this update,
        no matter what kind of update this is.

        Note:
            * Depending on the type of update and the user's 'Remain anonymous' setting, this
              could either be :class:`telegram.User`, :class:`telegram.Chat` or :obj:`None`.

        If no user whatsoever is associated with this update, this gives :obj:`None`. This
        is the case if any of

        * :attr:`poll`
        * :attr:`chat_boost`
        * :attr:`removed_chat_boost`
        * :attr:`message_reaction_count`
        * :attr:`deleted_business_messages`

        is present.

        Example:
            * If :attr:`message` is present, this will give either
              :attr:`telegram.Message.from_user` or :attr:`telegram.Message.sender_chat`.
            * If :attr:`poll_answer` is present, this will give either
              :attr:`telegram.PollAnswer.user` or :attr:`telegram.PollAnswer.voter_chat`.
            * If :attr:`channel_post` is present, this will give
              :attr:`telegram.Message.sender_chat`.

        .. versionadded:: 21.1
        """
        if self._effective_sender:
            return self._effective_sender

        sender: User | Chat | None = None

        if message := (
            self.message
            or self.edited_message
            or self.channel_post
            or self.edited_channel_post
            or self.business_message
            or self.edited_business_message
        ):
            sender = message.sender_chat

        elif self.poll_answer:
            sender = self.poll_answer.voter_chat

        elif self.message_reaction:
            sender = self.message_reaction.actor_chat

        if sender is None:
            sender = self.effective_user

        self._effective_sender = sender
        return sender

    @property
    def effective_chat(self) -> "Chat | None":
        """
        :class:`telegram.Chat`: The chat that this update was sent in, no matter what kind of
        update this is.
        If no chat is associated with this update, this gives :obj:`None`.
        This is the case, if :attr:`inline_query`,
        :attr:`chosen_inline_result`, :attr:`callback_query` from inline messages,
        :attr:`shipping_query`, :attr:`pre_checkout_query`, :attr:`poll`,
        :attr:`poll_answer`, :attr:`business_connection`, or :attr:`purchased_paid_media`
        is present.

        .. versionchanged:: 21.1
            This property now also considers :attr:`business_message`,
            :attr:`edited_business_message`, and :attr:`deleted_business_messages`.

        Example:
            If :attr:`message` is present, this will give :attr:`telegram.Message.chat`.

        """
        if self._effective_chat:
            return self._effective_chat

        chat = None

        if self.message:
            chat = self.message.chat

        elif self.edited_message:
            chat = self.edited_message.chat

        elif self.callback_query and self.callback_query.message:
            chat = self.callback_query.message.chat

        elif self.channel_post:
            chat = self.channel_post.chat

        elif self.edited_channel_post:
            chat = self.edited_channel_post.chat

        elif self.my_chat_member:
            chat = self.my_chat_member.chat

        elif self.chat_member:
            chat = self.chat_member.chat

        elif self.chat_join_request:
            chat = self.chat_join_request.chat

        elif self.chat_boost:
            chat = self.chat_boost.chat

        elif self.removed_chat_boost:
            chat = self.removed_chat_boost.chat

        elif self.message_reaction:
            chat = self.message_reaction.chat

        elif self.message_reaction_count:
            chat = self.message_reaction_count.chat

        elif self.business_message:
            chat = self.business_message.chat

        elif self.edited_business_message:
            chat = self.edited_business_message.chat

        elif self.deleted_business_messages:
            chat = self.deleted_business_messages.chat

        self._effective_chat = chat
        return chat

    @property
    def effective_message(self) -> Message | None:
        """
        :class:`telegram.Message`: The message included in this update, no matter what kind of
            update this is. More precisely, this will be the message contained in :attr:`message`,
            :attr:`edited_message`, :attr:`channel_post`, :attr:`edited_channel_post` or
            :attr:`callback_query` (i.e. :attr:`telegram.CallbackQuery.message`) or :obj:`None`, if
            none of those are present.

        .. versionchanged:: 21.1
            This property now also considers :attr:`business_message`, and
            :attr:`edited_business_message`.

        Tip:
            This property will only ever return objects of type :class:`telegram.Message` or
            :obj:`None`, never :class:`telegram.MaybeInaccessibleMessage` or
            :class:`telegram.InaccessibleMessage`.
            Currently, this is only relevant for :attr:`callback_query`, as
            :attr:`telegram.CallbackQuery.message` is the only attribute considered by this
            property that can be an object of these types.
        """
        if self._effective_message:
            return self._effective_message

        message: Message | None = None

        if self.message:
            message = self.message

        elif self.edited_message:
            message = self.edited_message

        elif self.callback_query:
            if (
                isinstance(cbq_message := self.callback_query.message, Message)
                or cbq_message is None
            ):
                message = cbq_message
            else:
                warn(
                    (
                        "`update.callback_query` is not `None`, but of type "
                        f"`{cbq_message.__class__.__name__}`. This is not considered by "
                        "`Update.effective_message`. Please manually access this attribute "
                        "if necessary."
                    ),
                    stacklevel=2,
                )

        elif self.channel_post:
            message = self.channel_post

        elif self.edited_channel_post:
            message = self.edited_channel_post

        elif self.business_message:
            message = self.business_message

        elif self.edited_business_message:
            message = self.edited_business_message

        self._effective_message = message
        return message

    @classmethod
    def de_json(cls, data: JSONDict, bot: "Bot | None" = None) -> "Update":
        """See :meth:`telegram.TelegramObject.de_json`."""
        data = cls._parse_data(data)

        data["message"] = de_json_optional(data.get("message"), Message, bot)
        data["edited_message"] = de_json_optional(data.get("edited_message"), Message, bot)
        data["inline_query"] = de_json_optional(data.get("inline_query"), InlineQuery, bot)
        data["chosen_inline_result"] = de_json_optional(
            data.get("chosen_inline_result"), ChosenInlineResult, bot
        )
        data["callback_query"] = de_json_optional(data.get("callback_query"), CallbackQuery, bot)
        data["shipping_query"] = de_json_optional(data.get("shipping_query"), ShippingQuery, bot)
        data["pre_checkout_query"] = de_json_optional(
            data.get("pre_checkout_query"), PreCheckoutQuery, bot
        )
        data["channel_post"] = de_json_optional(data.get("channel_post"), Message, bot)
        data["edited_channel_post"] = de_json_optional(
            data.get("edited_channel_post"), Message, bot
        )
        data["poll"] = de_json_optional(data.get("poll"), Poll, bot)
        data["poll_answer"] = de_json_optional(data.get("poll_answer"), PollAnswer, bot)
        data["my_chat_member"] = de_json_optional(
            data.get("my_chat_member"), ChatMemberUpdated, bot
        )
        data["chat_member"] = de_json_optional(data.get("chat_member"), ChatMemberUpdated, bot)
        data["chat_join_request"] = de_json_optional(
            data.get("chat_join_request"), ChatJoinRequest, bot
        )
        data["chat_boost"] = de_json_optional(data.get("chat_boost"), ChatBoostUpdated, bot)
        data["removed_chat_boost"] = de_json_optional(
            data.get("removed_chat_boost"), ChatBoostRemoved, bot
        )
        data["message_reaction"] = de_json_optional(
            data.get("message_reaction"), MessageReactionUpdated, bot
        )
        data["message_reaction_count"] = de_json_optional(
            data.get("message_reaction_count"), MessageReactionCountUpdated, bot
        )
        data["business_connection"] = de_json_optional(
            data.get("business_connection"), BusinessConnection, bot
        )
        data["business_message"] = de_json_optional(data.get("business_message"), Message, bot)
        data["edited_business_message"] = de_json_optional(
            data.get("edited_business_message"), Message, bot
        )
        data["deleted_business_messages"] = de_json_optional(
            data.get("deleted_business_messages"), BusinessMessagesDeleted, bot
        )
        data["purchased_paid_media"] = de_json_optional(
            data.get("purchased_paid_media"), PaidMediaPurchased, bot
        )

        return super().de_json(data=data, bot=bot)



#!/usr/bin/env python
# pylint: disable=unused-argument
# This program is dedicated to the public domain under the CC0 license.

"""
First, a few callback functions are defined. Then, those functions are passed to
the Application and registered at their respective places.
Then, the bot is started and runs until we press Ctrl-C on the command line.

Usage:
Example of a bot-user conversation using ConversationHandler.
Send /start to initiate the conversation.
Press Ctrl-C on the command line or send a signal to the process to stop the
bot.
"""

import logging

from telegram import ReplyKeyboardMarkup, ReplyKeyboardRemove, Update
from telegram.ext import (
    Application,
    CommandHandler,
    ContextTypes,
    ConversationHandler,
    MessageHandler,
    filters,
)

# Enable logging
logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
# set higher logging level for httpx to avoid all GET and POST requests being logged
logging.getLogger("httpx").setLevel(logging.WARNING)

logger = logging.getLogger(__name__)

CHOOSING, TYPING_REPLY, TYPING_CHOICE = range(3)

reply_keyboard = [
    ["Age", "Favourite colour"],
    ["Number of siblings", "Something else..."],
    ["Done"],
]
markup = ReplyKeyboardMarkup(reply_keyboard, one_time_keyboard=True)


def facts_to_str(user_data: dict[str, str]) -> str:
    """Helper function for formatting the gathered user info."""
    facts = [f"{key} - {value}" for key, value in user_data.items()]
    return "\n".join(facts).join(["\n", "\n"])


async def start(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    """Start the conversation and ask user for input."""
    await update.message.reply_text(
        "Hi! My name is Doctor Botter. I will hold a more complex conversation with you. "
        "Why don't you tell me something about yourself?",
        reply_markup=markup,
    )

    return CHOOSING


async def regular_choice(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    """Ask the user for info about the selected predefined choice."""
    text = update.message.text
    context.user_data["choice"] = text
    await update.message.reply_text(f"Your {text.lower()}? Yes, I would love to hear about that!")

    return TYPING_REPLY


async def custom_choice(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    """Ask the user for a description of a custom category."""
    await update.message.reply_text(
        'Alright, please send me the category first, for example "Most impressive skill"'
    )

    return TYPING_CHOICE


async def received_information(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    """Store info provided by user and ask for the next category."""
    user_data = context.user_data
    text = update.message.text
    category = user_data["choice"]
    user_data[category] = text
    del user_data["choice"]

    await update.message.reply_text(
        "Neat! Just so you know, this is what you already told me:"
        f"{facts_to_str(user_data)}You can tell me more, or change your opinion"
        " on something.",
        reply_markup=markup,
    )

    return CHOOSING


async def done(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    """Display the gathered info and end the conversation."""
    user_data = context.user_data
    if "choice" in user_data:
        del user_data["choice"]

    await update.message.reply_text(
        f"I learned these facts about you: {facts_to_str(user_data)}Until next time!",
        reply_markup=ReplyKeyboardRemove(),
    )

    user_data.clear()
    return ConversationHandler.END


def main() -> None:
    """Run the bot."""
    # Create the Application and pass it your bot's token.
    application = Application.builder().token("TOKEN").build()

    # Add conversation handler with the states CHOOSING, TYPING_CHOICE and TYPING_REPLY
    conv_handler = ConversationHandler(
        entry_points=[CommandHandler("start", start)],
        states={
            CHOOSING: [
                MessageHandler(
                    filters.Regex("^(Age|Favourite colour|Number of siblings)$"), regular_choice
                ),
                MessageHandler(filters.Regex("^Something else...$"), custom_choice),
            ],
            TYPING_CHOICE: [
                MessageHandler(
                    filters.TEXT & ~(filters.COMMAND | filters.Regex("^Done$")), regular_choice
                )
            ],
            TYPING_REPLY: [
                MessageHandler(
                    filters.TEXT & ~(filters.COMMAND | filters.Regex("^Done$")),
                    received_information,
                )
            ],
        },
        fallbacks=[MessageHandler(filters.Regex("^Done$"), done)],
    )

    application.add_handler(conv_handler)

    # Run the bot until the user presses Ctrl-C
    application.run_polling(allowed_updates=Update.ALL_TYPES)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
import pytest

from telegram import RefundedPayment
from tests.auxil.slots import mro_slots


@pytest.fixture(scope="module")
def refunded_payment():
    return RefundedPayment(
        RefundedPaymentTestBase.currency,
        RefundedPaymentTestBase.total_amount,
        RefundedPaymentTestBase.invoice_payload,
        RefundedPaymentTestBase.telegram_payment_charge_id,
        RefundedPaymentTestBase.provider_payment_charge_id,
    )


class RefundedPaymentTestBase:
    invoice_payload = "invoice_payload"
    currency = "EUR"
    total_amount = 100
    telegram_payment_charge_id = "telegram_payment_charge_id"
    provider_payment_charge_id = "provider_payment_charge_id"


class TestRefundedPaymentWithoutRequest(RefundedPaymentTestBase):
    def test_slot_behaviour(self, refunded_payment):
        inst = refunded_payment
        for attr in inst.__slots__:
            assert getattr(inst, attr, "err") != "err", f"got extra slot '{attr}'"
        assert len(mro_slots(inst)) == len(set(mro_slots(inst))), "duplicate slot"

    def test_de_json(self, offline_bot):
        json_dict = {
            "invoice_payload": self.invoice_payload,
            "currency": self.currency,
            "total_amount": self.total_amount,
            "telegram_payment_charge_id": self.telegram_payment_charge_id,
            "provider_payment_charge_id": self.provider_payment_charge_id,
        }
        refunded_payment = RefundedPayment.de_json(json_dict, offline_bot)
        assert refunded_payment.api_kwargs == {}

        assert refunded_payment.invoice_payload == self.invoice_payload
        assert refunded_payment.currency == self.currency
        assert refunded_payment.total_amount == self.total_amount
        assert refunded_payment.telegram_payment_charge_id == self.telegram_payment_charge_id
        assert refunded_payment.provider_payment_charge_id == self.provider_payment_charge_id

    def test_to_dict(self, refunded_payment):
        refunded_payment_dict = refunded_payment.to_dict()

        assert isinstance(refunded_payment_dict, dict)
        assert refunded_payment_dict["invoice_payload"] == refunded_payment.invoice_payload
        assert refunded_payment_dict["currency"] == refunded_payment.currency
        assert refunded_payment_dict["total_amount"] == refunded_payment.total_amount
        assert (
            refunded_payment_dict["telegram_payment_charge_id"]
            == refunded_payment.telegram_payment_charge_id
        )
        assert (
            refunded_payment_dict["provider_payment_charge_id"]
            == refunded_payment.provider_payment_charge_id
        )

    def test_equality(self):
        a = RefundedPayment(
            self.currency,
            self.total_amount,
            self.invoice_payload,
            self.telegram_payment_charge_id,
            self.provider_payment_charge_id,
        )
        b = RefundedPayment(
            self.currency,
            self.total_amount,
            self.invoice_payload,
            self.telegram_payment_charge_id,
            self.provider_payment_charge_id,
        )
        c = RefundedPayment("", 0, "", self.telegram_payment_charge_id)
        d = RefundedPayment(
            self.currency,
            self.total_amount,
            self.invoice_payload,
            "",
        )

        assert a == b
        assert hash(a) == hash(b)
        assert a is not b

        assert a == c
        assert hash(a) == hash(c)

        assert a != d
        assert hash(a) != hash(d)



import datetime as dtm
from collections.abc import Sequence
from typing import TypeAlias

from telegram import (
    AcceptedGiftTypes,
    BotCommand,
    BotDescription,
    BotName,
    BotShortDescription,
    BusinessBotRights,
    BusinessConnection,
    Chat,
    ChatAdministratorRights,
    ChatBoost,
    ChatBoostSource,
    ChatFullInfo,
    ChatInviteLink,
    ChatMember,
    File,
    ForumTopic,
    GameHighScore,
    Gift,
    Gifts,
    MenuButton,
    MessageId,
    OwnedGiftRegular,
    OwnedGifts,
    Poll,
    PollOption,
    PreparedInlineMessage,
    SentWebAppMessage,
    StarAmount,
    StarTransaction,
    StarTransactions,
    Sticker,
    StickerSet,
    Story,
    TelegramObject,
    Update,
    User,
    UserChatBoosts,
    UserProfileAudios,
    UserProfilePhotos,
    WebhookInfo,
)
from tests.auxil.build_messages import make_message

_DUMMY_USER = User(
    id=123456, is_bot=False, first_name="Dummy", last_name="User", username="dummy_user"
)
_DUMMY_DATE = dtm.datetime(1970, 1, 1, 0, 0, 0, 0, tzinfo=dtm.timezone.utc)
_DUMMY_STICKER = Sticker(
    file_id="dummy_file_id",
    file_unique_id="dummy_file_unique_id",
    width=1,
    height=1,
    is_animated=False,
    is_video=False,
    type="dummy_type",
)

_PREPARED_DUMMY_OBJECTS: dict[str, object] = {
    "bool": True,
    "BotCommand": BotCommand(command="dummy_command", description="dummy_description"),
    "BotDescription": BotDescription(description="dummy_description"),
    "BotName": BotName(name="dummy_name"),
    "BotShortDescription": BotShortDescription(short_description="dummy_short_description"),
    "BusinessConnection": BusinessConnection(
        user=_DUMMY_USER,
        id="123",
        user_chat_id=123456,
        date=_DUMMY_DATE,
        is_enabled=True,
        rights=BusinessBotRights(can_reply=True),
    ),
    "Chat": Chat(id=123456, type="dummy_type"),
    "ChatAdministratorRights": ChatAdministratorRights.all_rights(),
    "ChatFullInfo": ChatFullInfo(
        id=123456,
        type="dummy_type",
        accent_color_id=1,
        max_reaction_count=1,
        accepted_gift_types=AcceptedGiftTypes(
            unlimited_gifts=True,
            limited_gifts=True,
            unique_gifts=True,
            premium_subscription=True,
            gifts_from_channels=True,
        ),
    ),
    "ChatInviteLink": ChatInviteLink(
        "dummy_invite_link",
        creator=_DUMMY_USER,
        is_primary=True,
        is_revoked=False,
        creates_join_request=False,
    ),
    "ChatMember": ChatMember(user=_DUMMY_USER, status="dummy_status"),
    "File": File(file_id="dummy_file_id", file_unique_id="dummy_file_unique_id"),
    "ForumTopic": ForumTopic(message_thread_id=2, name="dummy_name", icon_color=1),
    "Gifts": Gifts(gifts=[Gift(id="dummy_id", sticker=_DUMMY_STICKER, star_count=1)]),
    "GameHighScore": GameHighScore(position=1, user=_DUMMY_USER, score=1),
    "int": 123456,
    "MenuButton": MenuButton(type="dummy_type"),
    "Message": make_message("dummy_text"),
    # Bad hack to get tests passing (we should not be using annotations as a key here)
    "Message | bool": make_message("dummy_text"),
    "MessageId": MessageId(123456),
    "OwnedGifts": OwnedGifts(
        total_count=1,
        gifts=[
            OwnedGiftRegular(
                gift=Gift(
                    id="id1",
                    sticker=Sticker(
                        "file_id", "file_unique_id", 512, 512, False, False, "regular"
                    ),
                    star_count=5,
                ),
                send_date=_DUMMY_DATE,
                owned_gift_id="some_id_1",
            )
        ],
    ),
    "Poll": Poll(
        id="dummy_id",
        question="dummy_question",
        options=[PollOption(text="dummy_text", voter_count=1)],
        is_closed=False,
        is_anonymous=False,
        total_voter_count=1,
        type="dummy_type",
        allows_multiple_answers=False,
    ),
    "PreparedInlineMessage": PreparedInlineMessage(id="dummy_id", expiration_date=_DUMMY_DATE),
    "SentWebAppMessage": SentWebAppMessage(inline_message_id="dummy_inline_message_id"),
    "StarAmount": StarAmount(amount=100, nanostar_amount=356),
    "StarTransactions": StarTransactions(
        transactions=[StarTransaction(id="dummy_id", amount=1, date=_DUMMY_DATE)]
    ),
    "Sticker": _DUMMY_STICKER,
    "StickerSet": StickerSet(
        name="dummy_name",
        title="dummy_title",
        stickers=[_DUMMY_STICKER],
        sticker_type="dummy_type",
    ),
    "Story": Story(chat=Chat(123, "prive"), id=123),
    "str": "dummy_string",
    "Update": Update(update_id=123456),
    "User": _DUMMY_USER,
    "UserChatBoosts": UserChatBoosts(
        boosts=[
            ChatBoost(
                boost_id="dummy_id",
                add_date=_DUMMY_DATE,
                expiration_date=_DUMMY_DATE,
                source=ChatBoostSource(source="dummy_source"),
            )
        ]
    ),
    "UserProfileAudios": UserProfileAudios(total_count=1, audios=[]),
    "UserProfilePhotos": UserProfilePhotos(total_count=1, photos=[[]]),
    "WebhookInfo": WebhookInfo(
        url="dummy_url",
        has_custom_certificate=False,
        pending_update_count=1,
    ),
}


def get_dummy_object(obj_type: type | str, as_tuple: bool = False) -> object:
    obj_type_name = obj_type.__name__ if isinstance(obj_type, type) else obj_type
    if (return_value := _PREPARED_DUMMY_OBJECTS.get(obj_type_name)) is None:
        raise ValueError(
            f"Dummy object of type '{obj_type_name}' not found. Please add it manually."
        )

    if as_tuple:
        return (return_value,)
    return return_value


_RETURN_TYPES: TypeAlias = bool | int | str | dict[str, object]
_RETURN_TYPE: TypeAlias = _RETURN_TYPES | tuple[_RETURN_TYPES, ...]


def _serialize_dummy_object(obj: object) -> _RETURN_TYPE:
    if isinstance(obj, Sequence) and not isinstance(obj, str):
        return tuple(_serialize_dummy_object(item) for item in obj)
    if isinstance(obj, str | int | bool):
        return obj
    if isinstance(obj, TelegramObject):
        return obj.to_dict()

    raise ValueError(f"Serialization of object of type '{type(obj)}' is not supported yet.")


def get_dummy_object_json_dict(obj_type: type | str, as_tuple: bool = False) -> _RETURN_TYPE:
    return _serialize_dummy_object(get_dummy_object(obj_type, as_tuple=as_tuple))



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
#  You should have received a copy of the GNU Lesser Public License
#  along with this program.  If not, see [http://www.gnu.org/licenses/].
"""This module contains the PollAnswerHandler class."""

from telegram import Update
from telegram.ext._handlers.basehandler import BaseHandler
from telegram.ext._utils.types import CCT, RT


class PollAnswerHandler(BaseHandler[Update, CCT, RT]):
    """Handler class to handle Telegram updates that contain a
    :attr:`poll answer <telegram.Update.poll_answer>`.

    Warning:
        When setting :paramref:`block` to :obj:`False`, you cannot rely on adding custom
        attributes to :class:`telegram.ext.CallbackContext`. See its docs for more info.

    Examples:
        :any:`Poll Bot <examples.pollbot>`

    Args:
        callback (:term:`coroutine function`): The callback function for this handler. Will be
            called when :meth:`check_update` has determined that an update should be processed by
            this handler. Callback signature::

                async def callback(update: Update, context: CallbackContext)

            The return value of the callback is usually ignored except for the special case of
            :class:`telegram.ext.ConversationHandler`.
        block (:obj:`bool`, optional): Determines whether the return value of the callback should
            be awaited before processing the next handler in
            :meth:`telegram.ext.Application.process_update`. Defaults to :obj:`True`.

            .. seealso:: :wiki:`Concurrency`

    Attributes:
        callback (:term:`coroutine function`): The callback function for this handler.
        block (:obj:`bool`): Determines whether the callback will run in a blocking way..

    """

    __slots__ = ()

    def check_update(self, update: object) -> bool:
        """Determines whether an update should be passed to this handler's :attr:`callback`.

        Args:
            update (:class:`telegram.Update` | :obj:`object`): Incoming update.

        Returns:
            :obj:`bool`

        """
        return isinstance(update, Update) and bool(update.poll_answer)



#!/usr/bin/env python
#
# A library that provides a Python interface to the Telegram Bot API
# Copyright (C) 2015-2026
# Leandro Toledo de Souza <devs@python-telegram-bot.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Lesser Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].
import asyncio
import datetime as dtm
import os
from pathlib import Path

import pytest

from telegram import (
    Bot,
    InputFile,
    MessageEntity,
    PhotoSize,
    ReplyParameters,
    Video,
    VideoQuality,
    Voice,
)
from telegram.constants import ParseMode
from telegram.error import BadRequest, TelegramError
from telegram.helpers import escape_markdown
from telegram.request import RequestData
from telegram.warnings import PTBDeprecationWarning
from tests.auxil.bot_method_checks import (
    check_defaults_handling,
    check_shortcut_call,
    check_shortcut_signature,
)
from tests.auxil.build_messages import make_message
from tests.auxil.files import data_file
from tests.auxil.slots import mro_slots


# Override `video` fixture to provide start_timestamp
@pytest.fixture(scope="module")
async def video(bot, chat_id):
    # The `video` object returned here does not have the `qualities` attribute.
    # Tests using actual VideoQuality objects from real Telegram responses
    # are implemented separately in `test_videoquality`.
    with data_file("telegram.mp4").open("rb") as f:
        return (
            await bot.send_video(
                chat_id, video=f, start_timestamp=VideoTestBase.start_timestamp, read_timeout=50
            )
        ).video


class VideoTestBase:
    width = 360
    height = 640
    duration = dtm.timedelta(seconds=5)
    file_size = 326534
    mime_type = "video/mp4"
    supports_streaming = True
    file_name = "telegram.mp4"
    start_timestamp = dtm.timedelta(seconds=3)
    cover = (PhotoSize("file_id", "unique_id", 640, 360, file_size=0),)
    qualities = (VideoQuality("video_file_id", "video_unique_id", 640, 360, "h264", file_size=0),)
    thumb_width = 180
    thumb_height = 320
    thumb_file_size = 1767
    caption = "<b>VideoTest</b> - *Caption*"
    video_file_url = "https://python-telegram-bot.org/static/testfiles/telegram.mp4"
    video_file_id = "5a3128a4d2a04750b5b58397f3b5e812"
    video_file_unique_id = "adc3145fd2e84d95b64d68eaa22aa33e"


class TestVideoWithoutRequest(VideoTestBase):
    def test_slot_behaviour(self, video):
        for attr in video.__slots__:
            assert getattr(video, attr, "err") != "err", f"got extra slot '{attr}'"
        assert len(mro_slots(video)) == len(set(mro_slots(video))), "duplicate slot"

    def test_creation(self, video):
        # Make sure file has been uploaded.
        assert isinstance(video, Video)
        assert isinstance(video.file_id, str)
        assert isinstance(video.file_unique_id, str)
        assert video.file_id
        assert video.file_unique_id

        assert isinstance(video.thumbnail, PhotoSize)
        assert isinstance(video.thumbnail.file_id, str)
        assert isinstance(video.thumbnail.file_unique_id, str)
        assert video.thumbnail.file_id
        assert video.thumbnail.file_unique_id

    def test_expected_values(self, video):
        assert video.width == self.width
        assert video.height == self.height
        assert video._duration == self.duration
        assert video.file_size == self.file_size
        assert video.mime_type == self.mime_type
        assert video._start_timestamp == self.start_timestamp

    def test_de_json(self, offline_bot):
        json_dict = {
            "file_id": self.video_file_id,
            "file_unique_id": self.video_file_unique_id,
            "width": self.width,
            "height": self.height,
            "duration": int(self.duration.total_seconds()),
            "mime_type": self.mime_type,
            "file_size": self.file_size,
            "file_name": self.file_name,
            "start_timestamp": int(self.start_timestamp.total_seconds()),
            "cover": [photo_size.to_dict() for photo_size in self.cover],
            "qualities": [video_quality.to_dict() for video_quality in self.qualities],
        }
        json_video = Video.de_json(json_dict, offline_bot)
        assert json_video.api_kwargs == {}

        assert json_video.file_id == self.video_file_id
        assert json_video.file_unique_id == self.video_file_unique_id
        assert json_video.width == self.width
        assert json_video.height == self.height
        assert json_video._duration == self.duration
        assert json_video.mime_type == self.mime_type
        assert json_video.file_size == self.file_size
        assert json_video.file_name == self.file_name
        assert json_video._start_timestamp == self.start_timestamp
        assert json_video.cover == self.cover
        assert json_video.qualities == self.qualities

    def test_to_dict(self, video):
        video_dict = video.to_dict()

        assert isinstance(video_dict, dict)
        assert video_dict["file_id"] == video.file_id
        assert video_dict["file_unique_id"] == video.file_unique_id
        assert video_dict["width"] == video.width
        assert video_dict["height"] == video.height
        assert video_dict["duration"] == int(self.duration.total_seconds())
        assert isinstance(video_dict["duration"], int)
        assert video_dict["mime_type"] == video.mime_type
        assert video_dict["file_size"] == video.file_size
        assert video_dict["file_name"] == video.file_name
        assert video_dict["start_timestamp"] == int(self.start_timestamp.total_seconds())
        assert isinstance(video_dict["start_timestamp"], int)

    def test_time_period_properties(self, PTB_TIMEDELTA, video):
        if PTB_TIMEDELTA:
            assert video.duration == self.duration
            assert isinstance(video.duration, dtm.timedelta)

            assert video.start_timestamp == self.start_timestamp
            assert isinstance(video.start_timestamp, dtm.timedelta)
        else:
            assert video.duration == int(self.duration.total_seconds())
            assert isinstance(video.duration, int)

            assert video.start_timestamp == int(self.start_timestamp.total_seconds())
            assert isinstance(video.start_timestamp, int)

    def test_time_period_int_deprecated(self, recwarn, PTB_TIMEDELTA, video):
        video.duration
        video.start_timestamp

        if PTB_TIMEDELTA:
            assert len(recwarn) == 0
        else:
            assert len(recwarn) == 2
            for i, attr in enumerate(["duration", "start_timestamp"]):
                assert f"`{attr}` will be of type `datetime.timedelta`" in str(recwarn[i].message)
                assert recwarn[i].category is PTBDeprecationWarning

    def test_equality(self, video):
        a = Video(video.file_id, video.file_unique_id, self.width, self.height, self.duration)
        b = Video("", video.file_unique_id, self.width, self.height, self.duration)
        c = Video(video.file_id, video.file_unique_id, 0, 0, 0)
        d = Video("", "", self.width, self.height, self.duration)
        e = Voice(video.file_id, video.file_unique_id, self.duration)

        assert a == b
        assert hash(a) == hash(b)
        assert a is not b

        assert a == c
        assert hash(a) == hash(c)

        assert a != d
        assert hash(a) != hash(d)

        assert a != e
        assert hash(a) != hash(e)

    async def test_error_without_required_args(self, offline_bot, chat_id):
        with pytest.raises(TypeError):
            await offline_bot.send_video(chat_id=chat_id)

    async def test_send_with_video(self, monkeypatch, offline_bot, chat_id, video):
        async def make_assertion(url, request_data: RequestData, *args, **kwargs):
            return request_data.json_parameters["video"] == video.file_id

        monkeypatch.setattr(offline_bot.request, "post", make_assertion)
        assert await offline_bot.send_video(chat_id, video=video)

    async def test_send_video_custom_filename(self, offline_bot, chat_id, video_file, monkeypatch):
        async def make_assertion(url, request_data: RequestData, *args, **kwargs):
            return next(iter(request_data.multipart_data.values()))[0] == "custom_filename"

        monkeypatch.setattr(offline_bot.request, "post", make_assertion)

        assert await offline_bot.send_video(chat_id, video_file, filename="custom_filename")

    @pytest.mark.parametrize("local_mode", [True, False])
    async def test_send_video_local_files(
        self, dummy_message_dict, monkeypatch, offline_bot, chat_id, local_mode
    ):
        try:
            offline_bot._local_mode = local_mode
            # For just test that the correct paths are passed as we have no local Bot API set up
            test_flag = False
            file = data_file("telegram.jpg")
            expected = file.as_uri()

            async def make_assertion(_, data, *args, **kwargs):
                nonlocal test_flag
                if local_mode:
                    test_flag = data.get("video") == expected and data.get("thumbnail") == expected
                else:
                    test_flag = isinstance(data.get("video"), InputFile) and isinstance(
                        data.get("thumbnail"), InputFile
                    )
                return dummy_message_dict

            monkeypatch.setattr(offline_bot, "_post", make_assertion)
            await offline_bot.send_video(chat_id, file, thumbnail=file)
            assert test_flag
        finally:
            offline_bot._local_mode = False

    async def test_get_file_instance_method(self, monkeypatch, video):
        async def make_assertion(*_, **kwargs):
            return kwargs["file_id"] == video.file_id

        assert check_shortcut_signature(Video.get_file, Bot.get_file, ["file_id"], [])
        assert await check_shortcut_call(video.get_file, video.get_bot(), "get_file")
        assert await check_defaults_handling(video.get_file, video.get_bot())

        monkeypatch.setattr(video.get_bot(), "get_file", make_assertion)
        assert await video.get_file()

    @pytest.mark.parametrize(
        ("default_bot", "custom"),
        [
            ({"parse_mode": ParseMode.HTML}, None),
            ({"parse_mode": ParseMode.HTML}, ParseMode.MARKDOWN_V2),
            ({"parse_mode": None}, ParseMode.MARKDOWN_V2),
        ],
        indirect=["default_bot"],
    )
    async def test_send_video_default_quote_parse_mode(
        self, default_bot, chat_id, video, custom, monkeypatch
    ):
        async def make_assertion(url, request_data: RequestData, *args, **kwargs):
            assert request_data.parameters["reply_parameters"].get("quote_parse_mode") == (
                custom or default_bot.defaults.quote_parse_mode
            )
            return make_message("dummy reply").to_dict()

        kwargs = {"message_id": 1}
        if custom is not None:
            kwargs["quote_parse_mode"] = custom

        monkeypatch.setattr(default_bot.request, "post", make_assertion)
        await default_bot.send_video(chat_id, video, reply_parameters=ReplyParameters(**kwargs))


class TestVideoWithRequest(VideoTestBase):
    @pytest.mark.parametrize("duration", [dtm.timedelta(seconds=5), 5])
    async def test_send_all_args(
        self, bot, chat_id, video_file, video, thumb_file, photo_file, duration
    ):
        message = await bot.send_video(
            chat_id,
            video_file,
            duration=duration,
            caption=self.caption,
            supports_streaming=self.supports_streaming,
            disable_notification=False,
            protect_content=True,
            width=video.width,
            height=video.height,
            parse_mode="Markdown",
            thumbnail=thumb_file,
            cover=photo_file,
            start_timestamp=self.start_timestamp,
            has_spoiler=True,
            show_caption_above_media=True,
        )

        assert isinstance(message.video, Video)
        assert isinstance(message.video.file_id, str)
        assert isinstance(message.video.file_unique_id, str)
        assert message.video.file_id
        assert message.video.file_unique_id
        assert message.video.width == video.width
        assert message.video.height == video.height
        assert message.video.duration == video.duration
        assert message.video.file_size == video.file_size

        assert message.caption == self.caption.replace("*", "")

        assert message.video.thumbnail.file_size == self.thumb_file_size
        assert message.video.thumbnail.width == self.thumb_width
        assert message.video.thumbnail.height == self.thumb_height

        assert message.video._start_timestamp == self.start_timestamp

        assert isinstance(message.video.cover, tuple)
        assert isinstance(message.video.cover[0], PhotoSize)

        assert message.video.file_name == self.file_name
        assert message.has_protected_content
        assert message.has_media_spoiler
        assert message.show_caption_above_media

    async def test_get_and_download(self, bot, video, chat_id, tmp_file):
        new_file = await bot.get_file(video.file_id)

        assert new_file.file_size == self.file_size
        assert new_file.file_unique_id == video.file_unique_id
        assert new_file.file_path.startswith("https://")

        await new_file.download_to_drive(tmp_file)

        assert tmp_file.is_file()

    async def test_send_mp4_file_url(self, bot, chat_id, video):
        message = await bot.send_video(chat_id, self.video_file_url, caption=self.caption)

        assert isinstance(message.video, Video)
        assert isinstance(message.video.file_id, str)
        assert isinstance(message.video.file_unique_id, str)
        assert message.video.file_id
        assert message.video.file_unique_id
        assert message.video.width == video.width
        assert message.video.height == video.height
        assert message.video.duration == video.duration
        assert message.video.file_size == video.file_size

        assert isinstance(message.video.thumbnail, PhotoSize)
        assert isinstance(message.video.thumbnail.file_id, str)
        assert isinstance(message.video.thumbnail.file_unique_id, str)
        assert message.video.thumbnail.file_id
        assert message.video.thumbnail.file_unique_id
        assert message.video.thumbnail.width == 51  # This seems odd that it's not self.thumb_width
        assert message.video.thumbnail.height == 90  # Ditto
        assert message.video.thumbnail.file_size == 645  # same

        assert message.caption == self.caption

    async def test_send_video_caption_entities(self, bot, chat_id, video):
        test_string = "Italic Bold Code"
        entities = [
            MessageEntity(MessageEntity.ITALIC, 0, 6),
            MessageEntity(MessageEntity.ITALIC, 7, 4),
            MessageEntity(MessageEntity.ITALIC, 12, 4),
        ]
        message = await bot.send_video(
            chat_id, video, caption=test_string, caption_entities=entities
        )

        assert message.caption == test_string
        assert message.caption_entities == tuple(entities)

    async def test_resend(self, bot, chat_id, video):
        message = await bot.send_video(chat_id, video.file_id)
        assert message.video == video

    @pytest.mark.parametrize("default_bot", [{"parse_mode": "Markdown"}], indirect=True)
    async def test_send_video_default_parse_mode_1(self, default_bot, chat_id, video):
        test_string = "Italic Bold Code"
        test_markdown_string = "_Italic_ *Bold* `Code`"

        message = await default_bot.send_video(chat_id, video, caption=test_markdown_string)
        assert message.caption_markdown == test_markdown_string
        assert message.caption == test_string

    @pytest.mark.parametrize("default_bot", [{"parse_mode": "Markdown"}], indirect=True)
    async def test_send_video_default_parse_mode_2(self, default_bot, chat_id, video):
        test_markdown_string = "_Italic_ *Bold* `Code`"

        message = await default_bot.send_video(
            chat_id, video, caption=test_markdown_string, parse_mode=None
        )
        assert message.caption == test_markdown_string
        assert message.caption_markdown == escape_markdown(test_markdown_string)

    @pytest.mark.parametrize("default_bot", [{"parse_mode": "Markdown"}], indirect=True)
    async def test_send_video_default_parse_mode_3(self, default_bot, chat_id, video):
        test_markdown_string = "_Italic_ *Bold* `Code`"

        message = await default_bot.send_video(
            chat_id, video, caption=test_markdown_string, parse_mode="HTML"
        )
        assert message.caption == test_markdown_string
        assert message.caption_markdown == escape_markdown(test_markdown_string)

    @pytest.mark.parametrize("default_bot", [{"protect_content": True}], indirect=True)
    async def test_send_video_default_protect_content(self, chat_id, default_bot, video):
        tasks = asyncio.gather(
            default_bot.send_video(chat_id, video),
            default_bot.send_video(chat_id, video, protect_content=False),
        )
        protected, unprotected = await tasks
        assert protected.has_protected_content
        assert not unprotected.has_protected_content

    @pytest.mark.parametrize(
        ("default_bot", "custom"),
        [
            ({"allow_sending_without_reply": True}, None),
            ({"allow_sending_without_reply": False}, None),
            ({"allow_sending_without_reply": False}, True),
        ],
        indirect=["default_bot"],
    )
    async def test_send_video_default_allow_sending_without_reply(
        self, default_bot, chat_id, video, custom
    ):
        reply_to_message = await default_bot.send_message(chat_id, "test")
        await reply_to_message.delete()
        if custom is not None:
            message = await default_bot.send_video(
                chat_id,
                video,
                allow_sending_without_reply=custom,
                reply_to_message_id=reply_to_message.message_id,
            )
            assert message.reply_to_message is None
        elif default_bot.defaults.allow_sending_without_reply:
            message = await default_bot.send_video(
                chat_id, video, reply_to_message_id=reply_to_message.message_id
            )
            assert message.reply_to_message is None
        else:
            with pytest.raises(BadRequest, match="Message to be replied not found"):
                await default_bot.send_video(
                    chat_id, video, reply_to_message_id=reply_to_message.message_id
                )

    async def test_error_send_empty_file(self, bot, chat_id):
        with Path(os.devnull).open("rb") as file, pytest.raises(TelegramError):
            await bot.send_video(chat_id, file)

    async def test_error_send_empty_file_id(self, bot, chat_id):
        with pytest.raises(TelegramError):
            await bot.send_video(chat_id, "")



# noqa: INP001
# pylint: disable=import-error
"""Configuration for the chango changelog tool"""

import re
from collections.abc import Collection
from pathlib import Path

from chango import Version
from chango.concrete import DirectoryChanGo, DirectoryVersionScanner, HeaderVersionHistory
from chango.concrete.sections import GitHubSectionChangeNote, Section, SectionVersionNote

version_scanner = DirectoryVersionScanner(base_directory=".", unreleased_directory="unreleased")


class ChangoSectionChangeNote(
    GitHubSectionChangeNote.with_sections(  # type: ignore[misc]
        [
            Section(uid="highlights", title="Highlights", sort_order=0),
            Section(uid="breaking", title="Breaking Changes", sort_order=1),
            Section(uid="security", title="Security Changes", sort_order=2),
            Section(uid="deprecations", title="Deprecations", sort_order=3),
            Section(uid="features", title="New Features", sort_order=4),
            Section(uid="bugfixes", title="Bug Fixes", sort_order=5),
            Section(uid="dependencies", title="Dependencies", sort_order=6),
            Section(uid="other", title="Other Changes", sort_order=7),
            Section(uid="documentation", title="Documentation", sort_order=8),
            Section(uid="internal", title="Internal Changes", sort_order=9),
        ]
    )
):
    """Custom change note type for PTB. Mainly overrides get_sections to map labels to sections"""

    OWNER = "python-telegram-bot"
    REPOSITORY = "python-telegram-bot"

    @classmethod
    def get_sections(
        cls,
        labels: Collection[str],
        issue_types: Collection[str] | None,
    ) -> set[str]:
        """Override get_sections to have customized auto-detection of relevant sections based on
        the pull request and linked issues. Certainly not perfect in all cases, but should be a
        good start for most PRs.
        """
        combined_labels = set(labels) | (set(issue_types or []))

        mapping = {
            "🐛 bug": "bugfixes",
            "💡 feature": "features",
            "🧹 chore": "internal",
            "⚙️ bot-api": "features",
            "⚙️ documentation": "documentation",
            "⚙️ tests": "internal",
            "⚙️ ci-cd": "internal",
            "⚙️ security": "security",
            "⚙️ examples": "documentation",
            "⚙️ type-hinting": "other",
            "🛠 refactor": "internal",
            "🛠 breaking": "breaking",
            "⚙️ dependencies": "dependencies",
            "🔗 github-actions": "internal",
            "🛠 code-quality": "internal",
        }

        # we want to return *all* from the mapping that are in the combined_labels
        # removing superfluous sections from the fragment is a tad easier than adding them
        found = {section for label, section in mapping.items() if label in combined_labels}

        # if we have not found any sections, we default to "other"
        return found or {"other"}


class CustomChango(DirectoryChanGo):
    """Custom ChanGo class for overriding release"""

    def release(self, version: Version) -> bool:
        """replace "14.5" with version.uid except in the contrib guide
        then call super
        """
        root = Path(__file__).parent.parent / "src"
        python_files = root.rglob("*.py")
        pattern = re.compile(r"NEXT\.VERSION")
        excluded_paths = {root / "docs/source/contribute.rst"}
        for file_path in python_files:
            if str(file_path) in excluded_paths:
                continue

            content = file_path.read_text(encoding="utf-8")
            modified = pattern.sub(version.uid, content)

            if content != modified:
                file_path.write_text(modified, encoding="utf-8")

        return super().release(version)


chango_instance = CustomChango(
    change_note_type=ChangoSectionChangeNote,
    version_note_type=SectionVersionNote,
    version_history_type=HeaderVersionHistory,
    scanner=version_scanner,
)
